from __future__ import absolute_import, division, print_function, unicode_literals

import itertools

from collections import deque, defaultdict, OrderedDict

import numpy as np
from numpy.core.umath_tests import inner1d

import networkx as nx

import chronic

from sklearn.cluster import DBSCAN

from ase import Atoms
from ase.visualize import view
from ase.data import covalent_radii

import matid.geometry
from matid.data import constants
from matid.core.linkedunits import LinkedUnitCollection, LinkedUnit
from matid.exceptions import SystaxError


class PeriodicFinder():
    """Used to find translationally periodic structures within atomic systems.
    """
    def __init__(
            self,
            angle_tol=constants.ANGLE_TOL,
            pos_tol_scaling=constants.POS_TOL_SCALING,
            cell_size_tol=constants.CELL_SIZE_TOL,
            max_2d_cell_height=constants.MAX_2D_CELL_HEIGHT,
            max_2d_single_cell_size=constants.MAX_SINGLE_CELL_SIZE,
            chem_similarity_threshold=constants.CHEM_SIMILARITY_THRESHOLD
        ):
        """
        Args:
            angle_tol(float): The angle below which vectors in the cell basis are
                considered to be parallel.
            pos_tol_factor(float): The factor for multiplying the position
                tolerance when searching neighbouring cell seed atoms.
            cell_size_tol(float): The tolerance for cell sizes to be considered
                equal. Given relative to the smallest cell size.
        """
        self.angle_tol = angle_tol
        self.pos_tol_scaling = pos_tol_scaling
        self.cell_size_tol = cell_size_tol
        self.max_2d_cell_height = max_2d_cell_height
        self.max_2d_single_cell_size = max_2d_single_cell_size,
        self.chem_similarity_threshold = chem_similarity_threshold

    def get_region(
            self,
            system,
            seed_index,
            max_cell_size,
            pos_tol,
            delaunay_threshold=None,
            bond_threshold=None,
            disp_tensor_mic=None,
            disp_factors=None,
            disp_tensor_finite=None,
            dist_matrix_radii_mic=None,
        ):
        """Tries to find the periodic regions, like surfaces, in an atomic
        system.

        Args:
            system(ase.Atoms): The system from which to find the periodic
                regions.
            seed_index(int): The index of the atom from which the search is
                initiated.
            max_cell_size(float): The maximum size of cell basis vectors.
            pos_tol(float): The tolerance that is allowed in the search. Given
                as absolute value in angstroms.

        Returns:
            linkedunitcollection or None: A LinkedUnitCollection object representing
                the region or None if no region could be identified.
        """
        if delaunay_threshold is None:
            delaunay_threshold = constants.DELAUNAY_THRESHOLD
        if bond_threshold is None:
            bond_threshold = constants.BOND_THRESHOLD

        # If the distance information is not given, calculate it here.
        pbc = system.get_pbc()
        if disp_tensor_mic is None and disp_factors is None and disp_tensor_finite is None and dist_matrix_radii_mic is None:
            pos = system.get_positions()
            cell = system.get_cell()
            pbc = system.get_pbc()
            disp_tensor_finite = matid.geometry.get_displacement_tensor(pos, pos)
            if pbc.any():
                disp_tensor_mic, disp_factors = matid.geometry.get_displacement_tensor(
                    pos,
                    pos,
                    cell,
                    pbc,
                    mic=True,
                    return_factors=True
                )
            else:
                disp_tensor_mic = disp_tensor_finite
                disp_factors = np.zeros(disp_tensor_finite.shape)
            dist_matrix_mic = np.linalg.norm(disp_tensor_mic, axis=2)

            # Calculate the distance matrix where the periodicity and the covalent
            # radii have been taken into account
            dist_matrix_radii_mic = np.array(dist_matrix_mic)
            num = system.get_atomic_numbers()
            radii = covalent_radii[num]
            radii_matrix = radii[:, None] + radii[None, :]
            dist_matrix_radii_mic -= radii_matrix

        self.disp_tensor_mic = disp_tensor_mic
        self.disp_tensor_finite = disp_tensor_finite
        self.disp_factors = disp_factors
        self.dist_matrix_radii_mic = dist_matrix_radii_mic
        self.pos_tol = pos_tol
        self.max_cell_size = max_cell_size
        region = None

        with chronic.Timer("find_possible_bases"):
            possible_spans, neighbour_mask, neighbour_factors = self._find_possible_bases(system, seed_index)
        with chronic.Timer("find_proto_cell"):
            proto_cell, offset, dim = self._find_proto_cell(
                system,
                seed_index,
                possible_spans,
                neighbour_mask,
                neighbour_factors,
                bond_threshold
            )

        # 1D is not handled
        if dim == 1 or proto_cell is None:
            return None

        # The indices of the periodic dimensions.
        periodic_indices = list(range(dim))

        # view(proto_cell)
        # print(proto_cell.get_cell())

        # Find a region that is spanned by the found unit cell

        with chronic.Timer("tracking"):
            unit_collection = self._find_periodic_region(
                system,
                dim == 2,
                delaunay_threshold,
                bond_threshold,
                seed_index,
                proto_cell,
                offset,
                periodic_indices,
            )

        with chronic.Timer("basis_indices"):
            i_indices = unit_collection.get_basis_indices()
            if len(i_indices) > 0:
                region = unit_collection
                region._pos_tol = pos_tol
                return region

    def _find_possible_bases(self, system, seed_index):
        """Finds all the possible vectors that might span a cell.

        Returns:
            np.ndarray: Array of found basis vectors
            np.ndarray: Mask that filters the neighbouring atoms
            np.ndarray: Factors that reveal in which periodic copy the
                neighbours are.
        """
        # Calculate a displacement tensor that takes into account the
        # periodicity of the system
        disp_tensor = self.disp_tensor_mic

        # If the search radius exceeds beyond the periodic boundaries, extend the system
        # Get the vectors that span from the seed to all other atoms
        seed_spans = disp_tensor[:, seed_index]
        atomic_numbers = system.get_atomic_numbers()

        # Find indices of atoms that are identical to seed atom
        seed_element = atomic_numbers[seed_index]
        identical_elem_mask = (atomic_numbers == seed_element)

        # Only keep spans that are smaller than the maximum vector length
        seed_span_lengths = np.linalg.norm(seed_spans, axis=1)
        distance_mask = (seed_span_lengths < self.max_cell_size)

        # Form a combined mask and filter spans with it
        combined_mask = (distance_mask) & (identical_elem_mask)
        combined_mask[seed_index] = False  # Ignore self
        bases = seed_spans[combined_mask]
        neighbour_factors = self.disp_factors[seed_index, distance_mask, :]

        return bases, distance_mask, neighbour_factors

    def _find_proto_cell(
            self,
            system,
            seed_index,
            possible_spans,
            neighbour_mask,
            neighbour_factors,
            bond_threshold
        ):
        """Used to find the best candidate for a unit cell basis that could
        generate a periodic region in the structure.

        Args:

        Returns:
            ase.Atoms: A system representing the best cell that was found
            np.ndarray: Position of the seed atom in the cell
        """
        positions = system.get_positions()
        numbers = system.get_atomic_numbers()

        # Find how many and which of the neighbouring atoms have a periodic
        # copy in the found directions
        with chronic.Timer("find_neighbour_copies"):
            neighbour_pos = positions[neighbour_mask]
            neighbour_num = numbers[neighbour_mask]
            neighbour_indices = np.where(neighbour_mask)[0]
            n_neighbours = len(neighbour_pos)
            n_spans = len(possible_spans)
            metric = np.empty((len(possible_spans)), dtype=int)
            adjacency_lists = []
            adjacency_lists_add = []
            adjacency_lists_sub = []
            neighbour_nodes = list(zip(neighbour_indices, neighbour_factors))

            for i_span, span in enumerate(possible_spans):

                # Calculate the scaled position tolerance
                i_pos_tol = np.full(n_neighbours, self.get_scaled_position_tolerance(span))

                i_adj_list = defaultdict(list)
                i_adj_list_add = defaultdict(list)
                i_adj_list_sub = defaultdict(list)
                add_pos = neighbour_pos + span
                sub_pos = neighbour_pos - span
                add_indices, _, _, add_factors = matid.geometry.get_matches(system, add_pos, neighbour_num, i_pos_tol)
                sub_indices, _, _, sub_factors = matid.geometry.get_matches(system, sub_pos, neighbour_num, i_pos_tol)

                n_metric = 0
                for i_neigh in range(n_neighbours):
                    i_add = add_indices[i_neigh]
                    i_sub = sub_indices[i_neigh]
                    i_add_factor = add_factors[i_neigh]
                    i_sub_factor = sub_factors[i_neigh]

                    if i_add is not None or i_sub is not None:
                        origin_factor = neighbour_factors[i_neigh]

                        if i_add is not None:
                            n_metric += 1
                            dest_factor = origin_factor + i_add_factor
                            i_adj_list[(neighbour_indices[i_neigh], tuple(origin_factor))].append((i_add, tuple(dest_factor)))
                            i_adj_list_add[(neighbour_indices[i_neigh], tuple(origin_factor))].append((i_add, tuple(dest_factor)))
                        if i_sub is not None:
                            n_metric += 1
                            dest_factor = origin_factor + i_sub_factor
                            i_adj_list[(neighbour_indices[i_neigh], tuple(origin_factor))].append((i_sub, tuple(dest_factor)))
                            i_adj_list_sub[(neighbour_indices[i_neigh], tuple(origin_factor))].append((i_sub, tuple(dest_factor)))

                metric[i_span] = n_metric
                adjacency_lists.append(i_adj_list)
                adjacency_lists_add.append(i_adj_list_add)
                adjacency_lists_sub.append(i_adj_list_sub)

        # Get the spans that come from the original cell basis if they are
        # smaller than the maximum cell size.
        with chronic.Timer("find_basis_spans"):
            periodic_spans = system.get_cell()
            periodic_span_lengths = np.linalg.norm(periodic_spans, axis=1)
            periodic_filter = periodic_span_lengths <= self.max_cell_size
            n_periodic_spans = periodic_filter.sum()
            if n_periodic_spans != 0:
                for i_per_span, per_span in enumerate(periodic_spans):
                    if periodic_filter[i_per_span]:

                        # Add the basis to the spans and add a full metric score
                        # for it.
                        possible_spans = np.concatenate((possible_spans, [per_span]), axis=0)
                        metric = np.concatenate((metric, [2*n_neighbours]), axis=0)

                        # Create the adjacency lists for the periodic span. There
                        # is full periodicity to neighbouring cells.
                        per_adjacency_list = defaultdict(list)
                        per_adjacency_list_add = defaultdict(list)
                        per_adjacency_list_sub = defaultdict(list)
                        i_factor = np.array((0, 0, 0))
                        i_factor[i_per_span] = 1
                        for i_neigh, neigh_factor in neighbour_nodes:

                            neigh_tuple = tuple(neigh_factor)
                            per_adjacency_list[(i_neigh, neigh_tuple)].append((i_neigh, tuple(neigh_factor + i_factor)))
                            per_adjacency_list[(i_neigh, neigh_tuple)].append((i_neigh, tuple(neigh_factor - i_factor)))
                            per_adjacency_list_add[(i_neigh, neigh_tuple)].append((i_neigh, tuple(neigh_factor + i_factor)))
                            per_adjacency_list_sub[(i_neigh, neigh_tuple)].append((i_neigh, tuple(neigh_factor - i_factor)))

                        adjacency_lists.append(per_adjacency_list)
                        adjacency_lists_add.append(per_adjacency_list_add)
                        adjacency_lists_sub.append(per_adjacency_list_sub)

        # Find the directions that repeat the neighbours above some preset
        # threshold. This is used to eliminate directions that are caused by
        # pure chance. The maximum score that a direction can get is
        # 2*n_neighbours. We specify that the score must be above 75% percent
        # of this maximum score to be considered a valid direction.
        valid_span_indices = np.where(metric >= 0.75*n_neighbours)[0]
        if len(valid_span_indices) == 0:
            return None, None, None

        # Find the best basis
        valid_span_metrics = metric[valid_span_indices]
        valid_spans = possible_spans[valid_span_indices]
        best_combo = self._find_best_basis(valid_spans, valid_span_metrics)
        dim = len(best_combo)

        # Currently 1D is not handled
        if dim == 1:
            return None, None, 1

        best_spans = valid_spans[best_combo]
        n_spans = len(best_spans)

        # Get the adjacency lists corresponding to the best spans
        with chronic.Timer("form_adjacency"):
            best_adjacency_lists = []
            best_adjacency_lists_add = []
            best_adjacency_lists_sub = []
            for i_span in best_combo:
                original_span_index = valid_span_indices[i_span]
                i_adjacency_list = adjacency_lists[original_span_index]
                i_adjacency_list_add = adjacency_lists_add[original_span_index]
                i_adjacency_list_sub = adjacency_lists_sub[original_span_index]
                best_adjacency_lists.append(i_adjacency_list)
                best_adjacency_lists_add.append(i_adjacency_list_add)
                best_adjacency_lists_sub.append(i_adjacency_list_sub)

        # Create a full periodicity graph for the found basis
        with chronic.Timer("form_periodicity_graph"):
            periodicity_graph_pbc = None
            full_adjacency_list_pbc = defaultdict(list)
            for i_adj in best_adjacency_lists:
                for key, value in i_adj.items():
                    full_adjacency_list_pbc[key].extend(value)
            periodicity_graph_pbc = nx.Graph(full_adjacency_list_pbc)

        # Expand the graph by exploring the links of the atoms that are in
        # neighbouring cells
        with chronic.Timer("form_periodicity_graph"):
            node_conn = {}
            for cell_node, neigh_fact in zip(neighbour_indices, neighbour_factors):
                for node in periodicity_graph_pbc.nodes():
                    node_index, node_factor = node
                    if node_index == cell_node and node_factor == tuple(neigh_fact):
                        connections = periodicity_graph_pbc[node]
                        disp = []
                        ns = []
                        for neighbour_index, neighbour_factor in connections:
                            i_disp = tuple(np.array(neighbour_factor) - np.array(node_factor))
                            disp.append(i_disp)
                            ns.append(neighbour_index)
                        node_conn[cell_node] = [ns, disp]

        new_graph = periodicity_graph_pbc.copy()
        for node in periodicity_graph_pbc.nodes():
            node_index, node_factor = node
            if node_factor != (0, 0, 0):
                i_conn = node_conn.get(node_index)
                if i_conn is None:
                    continue
                for i_ind, i_factor in zip(i_conn[0], i_conn[1]):
                    new_node = (i_ind, tuple(np.array(node_factor) + np.array(i_factor)))
                    if new_node not in new_graph:
                        new_graph.add_node(new_node)
                    if not new_graph.has_edge(node, new_node):
                        new_graph.add_edge(node, new_node)
        periodicity_graph_pbc = new_graph

        # import matplotlib.pyplot as plt
        # plt.subplot(111)
        # pos = nx.spring_layout(periodicity_graph_pbc)
        # nx.draw_networkx_nodes(periodicity_graph_pbc, pos)
        # nx.draw_networkx_edges(periodicity_graph_pbc, pos)
        # data = periodicity_graph_pbc.nodes(data=True)
        # labels = {x[0]: x[0] for x in data}
        # nx.draw_networkx_labels(periodicity_graph_pbc, pos, labels, font_size=16)
        # plt.show()

        # Get all disconnected subgraphs in the periodicity graph that takes
        # periodic boundaries into account
        graphs = [periodicity_graph_pbc.subgraph(c) for c in nx.connected_components(periodicity_graph_pbc)]

        # Filter out the basis atoms by checking how many were found with
        # respect to the number of copies of the seed atom. This equals to
        # checking that the size of the subgraph is similar to the size of the
        # graph where the seed atom is. This is needed because the number of
        # edges is not always sufficient when a lot is happening on the
        # surface.
        subgraph_size = []
        target_size = None
        for graph in graphs:
            nodes = graph.nodes()
            indices = set([node[0] for node in nodes])
            n_nodes = len(nodes)
            if seed_index in indices:
                target_size = n_nodes
            subgraph_size.append(n_nodes)
        temp_graphs = []
        for i_graph, graph_size in enumerate(subgraph_size):

            # Corresponds to the check \omega_v > 0.5n_{seed} in the article.
            if graph_size > 0.5*target_size:
                temp_graphs.append(graphs[i_graph])

        graphs = temp_graphs

        # Eliminate subgraphs that do not have enough periodicity.
        valid_graphs = []
        neighbourhood_set = set([(x[0], tuple(x[1])) for x in neighbour_nodes])
        for graph in graphs:

            # The periodicity is measured by the average degree of the nodes.
            # Only the degree of the nodes that are in the neighbourhood are
            # taken into account.
            degrees = []
            for node in graph.nodes():
                if node in neighbourhood_set:
                    i_degree = graph.degree(node)
                    degrees.append(i_degree)
            mean_degree = np.array(degrees).mean()

            # Corresponds to the check \omega_c > 2(d-1) in the article.
            if mean_degree > (dim-1)*2:
                valid_graphs.append(graph)

        # If no valid graphs found, no region can be tracked.
        if len(valid_graphs) == 0:
            return None, None, None

        # Each subgraph represents a group of atoms that repeat periodically in
        # each cell. Here we calculate a mean position of these atoms in the
        # cell.
        seed_nodes = None
        seed_group_index = None
        group_data_pbc = {
            "num": [],
            "ind": [],
            "nodes": []
        }
        # Determine the indices, nodes and numbers for each valid subgraph.
        # index_set = set()
        for i_graph, graph in enumerate(valid_graphs):
            nodes = graph.nodes()
            node_indices = [node[0] for node in nodes]

            if seed_index in set(node_indices):
                seed_group_index = i_graph
                seed_nodes = nodes

            group_data_pbc["ind"].append(node_indices)
            group_data_pbc["nodes"].append(nodes)
            group_data_pbc["num"].append(numbers[node_indices][0])

        # If the seed atom is not in a valid graph, no region could be found.
        if seed_group_index is None:
            return None, None, None

        if n_spans == 3:
            proto_cell, offset = self._find_proto_cell_3d(
                seed_index,
                seed_nodes,
                best_combo,
                best_spans,
                system,
                group_data_pbc,
                seed_group_index,
                best_adjacency_lists_add,
                best_adjacency_lists_sub,
            )
        elif n_spans == 2:

            # The seed group index can get updated by the cell search
            proto_cell, offset, seed_group_index = self._find_proto_cell_2d(
                seed_index,
                seed_nodes,
                best_combo,
                best_spans,
                system,
                group_data_pbc,
                seed_group_index,
                best_adjacency_lists_add,
                best_adjacency_lists_sub,
            )

            if proto_cell is None:
                return None, None, None

        two_valid_spans = n_spans == 2
        if n_spans == 3:
            # If the max_cell_size is bigger than an interlayer distance
            # between two 2D sheets, then a wrong cell with a lot of vacuum
            # might get detected. Here we check that the dimensionality of the
            # found 3D cell is correct.
            try:
                dimensionality = matid.geometry.get_dimensionality(proto_cell, bond_threshold)
            except SystaxError:
                return None, None, None
            if dimensionality != 3:
                # If the cell has three unit vectors, but does not exhibit the
                # correct dimensionality, try if two of the cell vectors are
                # still OK.
                if dimensionality == 2:
                    a_thickness = matid.geometry.get_thickness(proto_cell, 0)
                    b_thickness = matid.geometry.get_thickness(proto_cell, 1)
                    c_thickness = matid.geometry.get_thickness(proto_cell, 2)
                    reduced_dimension = np.argmin([a_thickness, b_thickness, c_thickness])
                    proto_cell = matid.geometry.get_minimized_cell(proto_cell, reduced_dimension, 2*self.pos_tol)
                    i_pbc = [True, True, True]
                    i_pbc[reduced_dimension] = False
                    cell_mask = [True, True, True]
                    cell_mask[reduced_dimension] = False
                    proto_cell.set_pbc(i_pbc)
                    best_combo = best_combo[cell_mask]
                    best_spans = best_spans[cell_mask]
                    two_valid_spans = True
                    n_spans = 2
                else:
                    return None, None, None

        if two_valid_spans:

            # If the best 2D vectors consists only of the simulation basis cell
            # vectors, check that these vectors are below a predefined size.
            # Otherwise the cell cannot be accepted because there is not enough
            # statistics about the cell contents to distinguish outliers.
            if n_periodic_spans > 0:
                periodic_span_indices = valid_span_indices[-n_periodic_spans:]
                best_span_ind = valid_span_indices[best_combo]
                if set(best_span_ind).issubset(set(periodic_span_indices)):
                    cell_lens = np.linalg.norm(best_spans, axis=1)
                    if np.any(cell_lens > self.max_2d_single_cell_size):
                        return None, None, None

            # Check the dimensionality
            dimensionality, cluster_labels = matid.geometry.get_dimensionality(proto_cell, bond_threshold, return_clusters=True)
            if dimensionality is None:
                # If the original system has more than one cluster, the system
                # has multiple stacked 2D sheets with identical periodicity. In
                # this case the unit cell should only comprise of atoms in the
                # cluster where the seed atom is in.
                for i_index, i_cluster in enumerate(cluster_labels):
                    try:
                        seed_group_index = i_cluster.index(seed_group_index)
                    except ValueError:
                        pass
                    else:
                        cluster_indices = i_cluster
                        break
                proto_cell = proto_cell[cluster_indices]

                # Retry to get the dimensionality for the cell in which the
                # cluster where the seed atom is in has been separated.
                # view(proto_cell)
                dimensionality = matid.geometry.get_dimensionality(proto_cell, bond_threshold)
                if dimensionality is None:
                    return None, None, None
                else:
                    if dimensionality != 2:
                        return None, None, None
            else:
                if dimensionality != 2:
                    return None, None, None

            # Check the cell thickness. 2D materials that are thicker than a
            # specified threshold are accepted.
            proto_cell = matid.geometry.get_minimized_cell(proto_cell, 2, 2*self.pos_tol)
            offset = proto_cell.get_positions()[seed_group_index]
            thickness = matid.geometry.get_thickness(proto_cell, 2)
            if thickness > self.max_2d_cell_height:
                return None, None, None

        return proto_cell, offset, n_spans

    def _find_proto_cell_3d(
            self,
            seed_index,
            seed_nodes,
            best_span_indices,
            best_spans,
            system,
            group_data_pbc,
            seed_group_index,
            adjacency_add,
            adjacency_sub,
        ):
        """Given a cell shape, this function is used to return a fully
        populated prototype unit cell for 3D systems.

        Args:
            seed_index(int): Index of the seed atom in the original system
            seed_nodes(list): List of tuples containing the node index as first
                entry, and the 3D index of the cell repetition as second entry.
                E.g. (0, (-1, 1, 2))
            best_span_indices():
            best_spans(np.ndarray): A set of basis vectors for the cell as 3x3
                array.
            system(ase.Atoms): Original system
            group_data_pbc():
            seed_group_index():
            adjacency_add():
            adjacency_sub():

        Returns:
            unit_cell(ase.Atoms): The unit cell
            offset(np.ndarray): The cartesian offset of the seed atom in the
                cell.
        """
        # Keep one occurrence for each seed index

        # Find the seed positions copies that are within the neighbourhood
        orig_cell = system.get_cell()

        # Find the cells in which the copies of the seed atom are at the
        # origin. Here we are reusing information from the displacement tensor
        # and the factors that tell in which periodic cell copy the match was
        # found originally.
        cells = np.zeros((len(seed_nodes), 3, 3))
        for i_node, node in enumerate(seed_nodes):

            node_index = node[0]
            node_factor = node[1]

            # multiplier: The direction in which the cell basis is searched. +1 or -1.
            # node_factor: The cell index of the starting node.
            # i_factor: The factor of the matched atom.

            # Handle each basis
            for i_basis in range(3):
                a_final_neighbour = None
                a_add = adjacency_add[i_basis][node]
                a_sub = adjacency_sub[i_basis][node]

                if a_add:
                    a_add_neighbour, i_add_factor = a_add[0]
                    if a_add_neighbour != node_index:
                        a_final_neighbour = a_add_neighbour
                        i_factor = i_add_factor
                        multiplier = 1
                elif a_sub:
                    a_sub_neighbour, i_sub_factor = a_sub[0]
                    if a_sub_neighbour != node_index:
                        a_final_neighbour = a_sub_neighbour
                        i_factor = i_sub_factor
                        multiplier = -1

                if a_final_neighbour is not None:
                    # Old version
                    # a_correction = np.dot((-np.array(node_factor) + np.array(i_factor)), orig_cell)
                    # a = multiplier*self.disp_tensor_finite[a_final_neighbour, node_index, :] + a_correction

                    # New version
                    a_correction = np.dot((-np.array(node_factor) + np.array(i_factor)), orig_cell)
                    a = self.disp_tensor_finite[a_final_neighbour, node_index, :] + a_correction
                    a *= multiplier

                else:
                    a = best_spans[i_basis, :]

                cells[i_node, i_basis, :] = a

        # Find the relative positions of atoms inside the cell. If for too many
        # cells atoms are found that do not belong to the basis, then the found
        # unit cell is not correct and a cell should not be returned.
        orig_pos = system.get_positions()
        inside_nodes = []
        inside_pos = []
        index_cell_map = {}
        for i_node, cell in zip(seed_nodes, cells):
            i_seed, i_seed_factor = i_node
            seed_pos = orig_pos[i_seed]

            if i_seed in index_cell_map:
                i_indices, i_pos, i_factors = index_cell_map[i_seed]
            else:
                i_indices, i_pos, i_factors = matid.geometry.get_positions_within_basis(
                    system,
                    cell,
                    seed_pos,
                    1e-8,
                    -1e-8,
                )

                # Here we ensure that the seed atom is included in the cell
                # positions. The search might miss the seed atom which is
                # exactly at the border of the cell
                # if i_seed not in i_indices:
                    # i_indices = np.append(i_indices, [i_seed], axis=0)
                    # i_pos = np.append(i_pos, np.array([[0, 0, 0.5]]), axis=0)
                    # i_factors = np.append(i_factors, [[0, 0, 0]], axis=0)

                index_cell_map[i_seed] = (i_indices, i_pos, i_factors)

            # if (i_seed == 6):
            # print(i_indices)

            # Add the seed node factor
            final_factors = []
            for factor in i_factors:
                i_final_factor = tuple(np.array(i_seed_factor) + factor)
                final_factors.append(i_final_factor)

            # Create nodes (index+factor) that were found
            i_cell_nodes = list(zip(i_indices, final_factors))
            inside_nodes.append(OrderedDict(zip(i_cell_nodes, range(len(i_cell_nodes)))))
            inside_pos.append(i_pos)

        # For each node in a network, find the first relative position. Wrap
        # and average these positions to get a robust final estimate.
        # averaged_rel_pos = np.zeros((len(group_data_pbc["ind"]), 3))
        averaged_rel_pos = []
        averaged_rel_num = []
        new_group_index = None

        for i_group, nodes in enumerate(group_data_pbc["nodes"]):
            group_num = group_data_pbc["num"][i_group]
            scaled_pos = []
            for group_node in nodes:
                for cell_nodes, cell_positions in zip(inside_nodes, inside_pos):
                    if group_node in cell_nodes:
                        pos_index = cell_nodes[group_node]
                        pos = cell_positions[pos_index]
                        scaled_pos.append(pos)
                        break

            # The basis location corresponding to this group is only added is
            # at least one occurrence is found in a cell.
            if len(scaled_pos) != 0:
                scaled_pos = np.array(scaled_pos)

                # Find the copy with minimum distance from origin
                distances = np.linalg.norm(scaled_pos, axis=1)
                min_dist_index = np.argmin(distances)
                min_dist_pos = scaled_pos[min_dist_index]

                # All the other copies are moved periodically to be near the
                # position that is closest to origin.
                distances = scaled_pos - min_dist_pos
                displacement = np.rint(distances)
                final_pos = scaled_pos - displacement

                # The average position is calculated
                group_avg = np.mean(final_pos, axis=0)
                averaged_rel_pos.append(group_avg)
                averaged_rel_num.append(group_num)

            if i_group == seed_group_index:
                new_group_index = len(averaged_rel_num) - 1
        seed_group_index = new_group_index

        averaged_rel_pos = np.array(averaged_rel_pos)

        proto_cell = Atoms(
            scaled_positions=averaged_rel_pos,
            symbols=averaged_rel_num,
            cell=best_spans,
            pbc=[True, True, True]
        )
        offset = proto_cell.get_positions()[seed_group_index]

        return proto_cell, offset

    def _find_proto_cell_2d(
            self,
            seed_index,
            seed_nodes,
            best_span_indices,
            best_spans,
            system,
            group_data_pbc,
            seed_group_index,
            adjacency_add,
            adjacency_sub,
        ):
        """Given a cell shape, this function is used to return a fully
        populated prototype unit cell for 2D systems.

        Args:
            seed_index(int): Index of the seed atom in the original system
            seed_nodes(list): List of tuples containing the node index as first
                entry, and the 3D index of the cell repetition as second entry.
                E.g. (0, (-1, 1, 2))
            best_span_indices():
            best_spans(np.ndarray): A set of basis vectors for the cell as 3x3
                array.
            system(ase.Atoms): Original system
            group_data_pbc():
            seed_group_index():
            adjacency_add():
            adjacency_sub():

        Returns:
            unit_cell(ase.Atoms): The unit cell
            offset(np.ndarray): The cartesian offset of the seed atom in the
                cell.
        """
        orig_cell = system.get_cell()

        # In 2D systems the maximum thickness of the system is defined by
        # max_cell_size.
        cutoff = self.max_cell_size

        # We need to make the third basis vector
        a = best_spans[0]
        b = best_spans[1]
        c = np.cross(a, b)
        c_norm = c/np.linalg.norm(c)
        c_norm = c_norm[None, :]
        c_test = 2*c_norm*cutoff

        basis = np.concatenate((best_spans, c_test), axis=0)

        # Find the cells in which the copies of the seed atom are at the
        # origin. Here we are reusing information from the displacement tensor
        # and the factors that tell in which periodic cell copy the match was
        # found originally.
        cells = np.zeros((len(seed_nodes), 3, 3))
        c_norms = np.zeros((len(seed_nodes), 3))
        for i_node, node in enumerate(seed_nodes):

            node_index = node[0]
            node_factor = node[1]

            # Handle each basis
            for i_basis in range(2):
                a_final_neighbour = None
                a_add = adjacency_add[i_basis][node]
                a_sub = adjacency_sub[i_basis][node]

                if a_add:
                    a_add_neighbour, i_add_factor = a_add[0]
                    if a_add_neighbour != node_index:
                        a_final_neighbour = a_add_neighbour
                        i_factor = i_add_factor
                        multiplier = 1
                elif a_sub:
                    a_sub_neighbour, i_sub_factor = a_sub[0]
                    if a_sub_neighbour != node_index:
                        a_final_neighbour = a_sub_neighbour
                        i_factor = i_sub_factor
                        multiplier = -1

                if a_final_neighbour is not None:
                    a_correction = np.dot((-np.array(node_factor) + np.array(i_factor)), orig_cell)
                    a = multiplier*self.disp_tensor_finite[a_final_neighbour, node_index, :] + a_correction
                else:
                    a = best_spans[i_basis, :]
                cells[i_node, i_basis, :] = a

            # Update the third axis for each cell.
            a = cells[i_node, 0]
            b = cells[i_node, 1]
            c = np.cross(a, b)
            c_norm = c/np.linalg.norm(c)
            c_norms[i_node, :] = c_norm
            c_norm = c_norm[None, :]
            c = 2*c_norm*cutoff
            cells[i_node, 2, :] = c

        # Find the relative positions of atoms inside the cell
        orig_pos = system.get_positions()
        inside_nodes = []
        inside_pos = []
        index_cell_map = {}

        for i_unit, (i_node, cell) in enumerate(zip(seed_nodes, cells)):
            i_seed, i_seed_factor = i_node
            seed_pos = orig_pos[i_seed]
            search_offset = -c_norms[i_unit, :]*cutoff
            search_coord = (seed_pos + search_offset)

            if i_seed in index_cell_map:
                i_indices, i_pos, i_factors = index_cell_map[i_seed]
            else:
                i_indices, i_pos, i_factors = matid.geometry.get_positions_within_basis(
                    system,
                    cell,
                    search_coord,
                    1e-8,
                    -1e-8,
                )

                # Here we ensure that the seed atom is included in the cell
                # positions. The search might miss the seed atom which is
                # exactly at the border of the cell
                # if i_seed not in i_indices:
                    # i_indices = np.append(i_indices, [i_seed], axis=0)
                    # i_pos = np.append(i_pos, np.array([[0, 0, 0.5]]), axis=0)
                    # i_factors = np.append(i_factors, [[0, 0, 0]], axis=0)

                index_cell_map[i_seed] = (i_indices, i_pos, i_factors)

            # Add the seed node factor
            final_factors = []
            for factor in i_factors:
                i_final_factor = tuple(np.array(i_seed_factor) + factor)
                final_factors.append(i_final_factor)

            # Create nodes (index+factor) that were found
            i_cell_nodes = list(zip(i_indices, final_factors))
            inside_nodes.append(OrderedDict(zip(i_cell_nodes, range(len(i_cell_nodes)))))
            inside_pos.append(i_pos)

        # For each node in a network, find the first relative position. Wrap
        # and average these positions to get a robust final estimate.
        averaged_rel_pos = []
        averaged_rel_num = []
        new_group_index = None
        for i_group, nodes in enumerate(group_data_pbc["nodes"]):
            scaled_pos = []
            group_num = group_data_pbc["num"][i_group]

            for group_node in nodes:
                for cell_nodes, cell_positions in zip(inside_nodes, inside_pos):
                    if group_node in cell_nodes:
                        pos_index = cell_nodes[group_node]
                        pos = cell_positions[pos_index]
                        scaled_pos.append(pos)
                        break

            # The basis location corresponding to this group is only added is
            # at least one occurrence is found in a cell.
            if len(scaled_pos) != 0:
                scaled_pos = np.array(scaled_pos)

                # Only wrap the 2D positions
                scaled_pos_2d = scaled_pos[:, 0:2]
                scaled_pos_2d %= 1

                # Find the copy with minimum distance from origin
                distances = np.linalg.norm(scaled_pos_2d, axis=1)
                min_dist_index = np.argmin(distances)
                min_dist_pos = scaled_pos_2d[min_dist_index]

                # All the other copies are moved periodically to be near the
                # position that is closest to origin.
                distances = scaled_pos_2d - min_dist_pos
                displacement = np.rint(distances)
                final_pos_2d = scaled_pos_2d - displacement

                # The average position is calculated
                scaled_pos[:, 0:2] = final_pos_2d
                group_avg = np.mean(scaled_pos, axis=0)
                averaged_rel_pos.append(group_avg)
                averaged_rel_num.append(group_num)

            if i_group == seed_group_index:
                new_group_index = len(averaged_rel_num) - 1
        seed_group_index = new_group_index

        # Move the seed positions back to the origin now that the search has
        # been performed
        averaged_rel_pos = np.array(averaged_rel_pos)

        proto_cell = Atoms(
            cell=basis,
            scaled_positions=averaged_rel_pos,
            symbols=averaged_rel_num,
            pbc=[True, True, False]
        )
        offset = proto_cell.get_positions()[seed_group_index]

        return proto_cell, offset, seed_group_index

    def _find_best_basis(self, valid_spans, valid_span_metrics):
        """Used to choose the best basis from a set of valid ones.

        The given set of basis vectors should represent ones that reproduce the
        correct periodicity. This function then chooses one with correct
        dimensionality, minimal volume and maximal orthogonality.

        Args:
            valid_spans(np.ndarray):
            valid_span_metrics(np.ndarray):

        Returns:
            np.ndarray: Indices of the best spans from the given list of valid
            spans.
        """
        # Normed spans
        norms = np.linalg.norm(valid_spans, axis=1)
        norm_spans = valid_spans / norms[:, None]
        n_spans = len(valid_spans)

        if n_spans == 1:
            return [0]
        elif n_spans == 2:
            best_indices = self._find_best_2d_basis(norm_spans, norms, valid_span_metrics)
            return best_indices

        # The angle threshold for validating cells
        angle_threshold = np.pi/180*self.angle_tol
        angle_thres_sin = abs(np.sin(angle_threshold))

        # Create combinations of normed spans
        span_indices = range(len(valid_spans))
        combo_indices = np.array(list(itertools.combinations(span_indices, 3)))
        normed_combos = norm_spans[combo_indices]

        # Create arrays containing the three angles for each combination. The
        # normalized zero vectors will have values NaN
        n_jk = np.cross(normed_combos[:, 1, :], normed_combos[:, 2, :])
        n_jk_norm = np.linalg.norm(n_jk, axis=1)
        with np.errstate(divide='ignore', invalid='ignore'):
            n_jk_hat = n_jk / n_jk_norm[:, None]
        alpha_ijk = np.abs(inner1d(normed_combos[:, 0, :], n_jk_hat))
        n_ki = np.cross(normed_combos[:, 2, :], normed_combos[:, 0, :])
        n_ki_norm = np.linalg.norm(n_ki, axis=1)
        with np.errstate(divide='ignore', invalid='ignore'):
            n_ki_hat = n_ki / n_ki_norm[:, None]
        alpha_jki = np.abs(inner1d(normed_combos[:, 1, :], n_ki_hat))
        n_ij = np.cross(normed_combos[:, 0, :], normed_combos[:, 1, :])
        n_ij_norm = np.linalg.norm(n_ij, axis=1)
        with np.errstate(divide='ignore', invalid='ignore'):
            n_ij_hat = n_ij / n_ij_norm[:, None]
        alpha_kij = np.abs(inner1d(normed_combos[:, 2, :], n_ij_hat))

        # The angles alpha, beta and gamma are cos(x), where x is the angle
        # between the vector and the normal of the plane. To get the angle y
        # between vector and plane, we use cos(x) = cos(pi/2 - y) = sin(y)
        with np.errstate(invalid='ignore'):
            ijk_mask = alpha_ijk >= angle_thres_sin
            jki_mask = alpha_jki >= angle_thres_sin
            kij_mask = alpha_kij >= angle_thres_sin
        angles_mask = ijk_mask & jki_mask & kij_mask

        # Number of valid angles for each combination
        n_valids = np.sum(angles_mask)

        # If there are three angles that are above the threshold, the cell is 3D
        if n_valids > 0:

            valid_indices = combo_indices[angles_mask]

            # Filter out combos that do not have metric score close to maximum
            # score that was found. This step is needed to filter out invalid
            metrics = valid_span_metrics[valid_indices]
            metric_sum = np.sum(metrics, axis=1)
            max_metric = metric_sum.max()
            metric_filter = metric_sum == max_metric
            valid_indices = valid_indices[metric_filter]

            # Filter the set into group with volume closest to the smallest
            # that was found
            set_norms = norms[valid_indices]
            set_volumes = alpha_ijk[angles_mask][metric_filter]*n_jk_norm[angles_mask][metric_filter]*np.prod(set_norms, axis=1)
            smallest_volume = set_volumes.min()
            smallest_cell_filter = set_volumes <= (1+self.cell_size_tol)*smallest_volume
            valid_indices = valid_indices[smallest_cell_filter]

            # From the group with smallest volume find a combination with
            # highest orthogonality
            n_ij_filtered = n_ij[angles_mask][metric_filter][smallest_cell_filter]
            n_ki_filtered = n_ki[angles_mask][metric_filter][smallest_cell_filter]
            n_jk_filtered = n_jk[angles_mask][metric_filter][smallest_cell_filter]
            ortho = 3 - inner1d(n_ij_filtered, n_ij_filtered) - inner1d(n_ki_filtered, n_ki_filtered) - inner1d(n_jk_filtered, n_jk_filtered)
            max_ortho_filter = np.argmin(ortho)
            best_span_indices = valid_indices[max_ortho_filter]

            # OLD VERSION
            # angle_sum = alpha_ijk[angles_mask] + alpha_jki[angles_mask] + alpha_kij[angles_mask]
            # angle_set = angle_sum[metric_filter][smallest_cell_filter]
            # biggest_angle_sum_filter = np.argmax(angle_set)
            # best_span_indices = valid_indices[biggest_angle_sum_filter]

        else:
            best_span_indices = self._find_best_2d_basis(norm_spans, norms, valid_span_metrics)

        return best_span_indices

    def _find_best_2d_basis(self, norm_spans, norms, valid_span_metrics):
        """Used to find the best 2D basis for a set of vectors.
        """
        # The angle threshold for validating cells
        angle_threshold = np.pi/180*self.angle_tol  # 10 degrees
        angle_thres_sin = np.sin(angle_threshold)

        # Calculate the cross-product between all normed pairs.
        n_spans = len(norm_spans)
        pair_crosses = np.zeros((n_spans, n_spans, 3))
        for i in range(n_spans):
            for j in range(n_spans):
                if j > i:
                    cross = np.cross(norm_spans[i], norm_spans[j])
                    pair_crosses[i, j, :] = cross
                    pair_crosses[j, i, :] = -cross

        # Get all pairs that have angle bigger than threshold
        up_indices = np.triu_indices(n_spans, k=1)
        valid_indices_a = up_indices[0]
        valid_indices_b = up_indices[1]
        valid_indices = np.concatenate((valid_indices_a[:, None], valid_indices_b[:, None]), axis=1)
        crosses = pair_crosses[up_indices]
        sin_angle = np.abs(np.linalg.norm(crosses, axis=1))
        angle_filter = sin_angle >= angle_thres_sin
        valid_indices = valid_indices[angle_filter]
        n_cross_valids = len(valid_indices)

        # For 2D structures the best span pair has lowest area, angle
        # closest to 90 and sum of metric close to maximum that was found
        if n_cross_valids > 0:

            # Get all pairs that have metric close to maximum
            metrics = valid_span_metrics[valid_indices]
            metric_sum = np.sum(metrics, axis=1)
            max_metric = metric_sum.max()
            metric_filter = metric_sum == max_metric
            valid_indices = valid_indices[metric_filter]

            # Find group of cells by finding cells with smallest area
            crosses = pair_crosses[valid_indices[:, 0], valid_indices[:, 1]]
            sin_angle = np.abs(np.linalg.norm(crosses, axis=1))
            valid_norms = np.prod(norms[valid_indices], axis=1)
            areas = valid_norms*sin_angle
            smallest_area = areas.min()
            smallest_cells_filter = areas < (1+self.cell_size_tol)*smallest_area
            valid_indices = valid_indices[smallest_cells_filter]

            # From the group with smallest area find a combination with
            # highest orthogonality
            crosses = pair_crosses[valid_indices[:, 0], valid_indices[:, 1]]
            sin_angle = np.abs(np.linalg.norm(crosses, axis=1))
            biggest_angle_index = np.argmax(sin_angle)

            best_span_indices = valid_indices[biggest_angle_index]

        # For 1D structures the best span is the shortest one
        else:
            best_span_indices = np.array([np.argmin(norms)])

        return best_span_indices

    def _find_periodic_region(
            self,
            system,
            is_2d,
            tesselation_distance,
            bond_threshold,
            seed_index,
            unit_cell,
            seed_position,
            periodic_indices,
        ):
        """Used to find atoms that are generated by a given unit cell and a
        given origin.

        Args:
            system(ASE.Atoms): Original system from which the periodic
                region is searched
            is_2d(boolean): Is the system a 2D material with cells in only one plane
            seed_index(int): Index of the atom from which the search is started
            unit_cell(ASE.Atoms): Repeating unit from which the searched region
                is composed of
            seed_position(np.ndarray): Cartesian position of the seed atom with
                respect to the unit cell origin.
            periodic_indices(sequence of int): Indices of the basis vectors
                that are periodic
            flat_indices(sequence of int): Indices of the basis vectors
                that are nearly zero vectors.
        """
        positions = system.get_positions()
        atomic_numbers = system.get_atomic_numbers()
        seed_pos = positions[seed_index]
        seed_number = atomic_numbers[seed_index]

        # Create a map between an atomic number and indices in the system
        number_to_index_map = {}
        number_to_pos_map = {}
        atomic_number_set = set(atomic_numbers)
        for number in atomic_number_set:
            number_indices = np.where(atomic_numbers == number)[0]
            number_to_index_map[number] = number_indices
            number_to_pos_map[number] = positions[number_indices]

        searched_cell_indices = set()
        used_indices = set()
        used_seed_indices = set()
        searched_vacancy_positions = []
        queue = deque()
        collection = LinkedUnitCollection(
            system,
            unit_cell,
            is_2d,
            self.dist_matrix_radii_mic,
            self.disp_tensor_finite,
            tesselation_distance,
            self.chem_similarity_threshold,
            bond_threshold,
        )
        multipliers = self._get_multipliers(periodic_indices)

        # Start off the queue
        self._find_region_rec(
            system,
            collection,
            number_to_index_map,
            number_to_pos_map,
            seed_index,
            seed_pos,
            seed_number,
            unit_cell,
            seed_position,
            searched_cell_indices,
            (0, 0, 0),
            used_indices,
            searched_vacancy_positions,
            periodic_indices,
            queue,
            multipliers,
            used_seed_indices)

        # Keep searching while new cells are found
        finished = False
        while not finished:
            try:
                queue_seed_index, queue_seed_pos, queue_index, queue_cell = queue.popleft()
            except IndexError:
                finished = True
            else:
                self._find_region_rec(
                    system,
                    collection,
                    number_to_index_map,
                    number_to_pos_map,
                    queue_seed_index,
                    queue_seed_pos,
                    seed_number,
                    queue_cell,
                    seed_position,
                    searched_cell_indices,
                    queue_index,
                    used_indices,
                    searched_vacancy_positions,
                    periodic_indices,
                    queue,
                    multipliers,
                    used_seed_indices)

        return collection

    def _get_multipliers(self, periodic_indices):
        """Used to calculate the multipliers that are used to multiply the cell
        basis vectors to find new unit cells.
        """
        # Here we decide the new seed points where the search is extended.
        n_periodic_dim = len(periodic_indices)
        if n_periodic_dim == 3:
            multipliers = np.array([
                [1, 0, 0],
                [0, 1, 0],
                [0, 0, 1],
                [-1, 0, 0],
                [0, -1, 0],
                [0, 0, -1],
            ])

        if n_periodic_dim == 2:
            multipliers = np.array([
                [1, 0, 0],
                [0, 1, 0],
                [-1, 0, 0],
                [0, -1, 0],
            ])

        return multipliers

    def _find_region_rec(
            self,
            system,
            collection,
            number_to_index_map,
            number_to_pos_map,
            seed_index,
            seed_pos,
            seed_atomic_number,
            unit_cell,
            seed_offset,
            searched_cell_indices,
            cell_index,
            used_indices,
            searched_vacancy_positions,
            periodic_indices,
            queue,
            multipliers,
            used_seed_indices
        ):
        """
        Args:
            system(ASE.Atoms): The original system from which the periodic
                region is searched.
            collection(LinkedUnitCollection): Container for LinkedUnits that
                belong to a found region.
            number_to_index_map(dict): Connects atomic number to indices in the
                original system.
            number_to_pos_map(dict): Connects atomic number to positions in the
                original system.
            seed_index(int): Index of the seed atom in the original system.
            seed_pos(np.ndarray): Position of the seed atom in cartesian coordinates.
            seed_atomic_number(int): Atomic number of the seed atom.
            unit_cell(ASE.Atoms): The current guess for the unit cell.
            seed_offset(np.ndrray): Cartesian offset of the seed atom from the unit cell
                origin.
            searched_cell_indices(set): Set of 3D indices that have been searched.
            cell_index(tuple): The 3D coordinate of this unit cell.
            used_seed_indices(set): The indices that have been used as seeds.
            periodic_indices(sequence of int): The indices of the basis vectors
                that are periodic
        """
        # Check if this cell has already been searched
        if tuple(cell_index) in searched_cell_indices:
            return
        else:
            searched_cell_indices.add(tuple(cell_index))

        # Try to get the scaled positions for atoms in this new cell. If the
        # cell is non-invertible, then this cell is not processed.
        try:
            # Wrapping is here disabled because it does not handle well values
            # that are negative within machine precision.
            cell_pos = unit_cell.get_scaled_positions(wrap=False)
        except:
            return
        cell_num = unit_cell.get_atomic_numbers()
        old_basis = unit_cell.get_cell()

        new_seed_indices = []
        new_seed_pos = []
        new_cell_indices = []
        orig_cell = system.get_cell()
        orig_pbc = system.get_pbc()

        # Translate and wrap the searched positions
        test_pos = unit_cell.get_positions() - seed_offset + seed_pos
        test_pos = matid.geometry.to_scaled(orig_cell, test_pos, pbc=orig_pbc, wrap=True)
        test_pos = matid.geometry.to_cartesian(orig_cell, test_pos)

        # Find the atoms that match the positions in the original basis
        disps = unit_cell.get_positions() - seed_offset
        pos_tolerances = self.get_scaled_position_tolerance(disps)

        matches, substitutions, vacancies, _ = matid.geometry.get_matches(
            system,
            test_pos,
            cell_num,
            pos_tolerances,
        )

        # Associate the matched atoms to this cell
        for match in matches:
            if match is not None:
                collection._index_cell_map[match] = cell_index

        # Add all the matches into the lists containing already searched
        # locations.
        used_indices.update(matches)

        # Only allow substitutions that have not been added already.
        valid_substitutions = []
        new_subst = []
        for subst in substitutions:
            true_subst = None
            if subst is not None:
                subst_ind = subst.index
                if subst_ind not in used_indices:
                    true_subst = subst
                new_subst.append(subst_ind)
            valid_substitutions.append(true_subst)
        used_indices.update(new_subst)

        # Correct the vacancy positions by the seed pos, seed offset and cell
        # periodicity. Add the vacancy only if it has not already been added
        # before.
        new_vacancy_pos = []
        valid_vacancies = []
        vacancy_pos_array = np.array(searched_vacancy_positions)
        for vacancy in vacancies:
            # Check if this vacancy has already been found
            if len(searched_vacancy_positions) != 0:
                vac_dist = np.linalg.norm(vacancy_pos_array - vacancy.position, axis=1)
                if vac_dist.min() > self.pos_tol:
                    new_vacancy_pos.append(vacancy.position)
                    valid_vacancies.append(vacancy)
            else:
                new_vacancy_pos.append(vacancy.position)
                valid_vacancies.append(vacancy)
        searched_vacancy_positions.extend(new_vacancy_pos)

        # Find the neighbouring cells for extending the search
        dislocations = np.dot(multipliers, old_basis)
        new_cell, new_seed_indices, new_seed_pos, new_cell_indices = self._find_new_seeds_and_cell(
            system,
            seed_index,
            seed_pos,
            seed_atomic_number,
            dislocations,
            multipliers,
            old_basis,
            used_indices,
            cell_index,
            searched_cell_indices,
            used_seed_indices,
            collection._used_points,
            collection._search_graph,
            collection._index_cell_map
        )

        # If there are matches or substitutional atoms in the unit, add them to
        # the collection
        new_unit = LinkedUnit(cell_index, seed_index, seed_pos, new_cell, matches, valid_substitutions, valid_vacancies)
        collection[cell_index] = new_unit

        # Save the updated cell shape for the new cells in the queue
        new_sys = Atoms(
            cell=new_cell,
            scaled_positions=cell_pos,
            symbols=cell_num,
            pbc=unit_cell.get_pbc()
        )
        cells = len(new_seed_pos)*[new_sys]

        # Add the found neighbours to a queue
        queue.extend(list(zip(new_seed_indices, new_seed_pos, new_cell_indices, cells)))

    def get_scaled_position_tolerance(self, displacements):
        """Used to calculate the position tolerance that is scaled by the
        search distance.
        """
        # Add new axis to sigle vector displacements
        if len(displacements.shape) == 1:
            displacements = displacements[None, :]
        distance = np.linalg.norm(displacements, axis=1)
        scaled_tol = (1+self.pos_tol_scaling*distance)*self.pos_tol

        return scaled_tol

    def _find_new_seeds_and_cell(
            self,
            system,
            seed_index,
            seed_pos,
            seed_atomic_number,
            dislocations,
            multipliers,
            old_cell,
            used_indices,
            cell_index,
            searched_cell_indices,
            used_seed_indices,
            used_points,
            search_graph,
            index_cell_map
        ):
        """When given a prototype unit cell shape and a set of search
        directions, searches for new seed atoms that are used to initiate a
        search for a new repetition for a unit cell.

        Args:
            system(ase.Atoms): The system from which the seed atoms are
                searched.
            seed_index(int): The index of the atom from which the search is
                started.
            seed_pos(np.ndarray): The position vector of the seed atom.
            seed_atomic_number(int): The atomic number of the seed atom.
            dislocation(np.ndarray): An array of dislocation vectors given
                relative to the seed position.
            multipliers(np.ndarray): Multiplications of the unit cell
                corresponding to the given dislocation vectors.
            old_cell(np.ndarray): The unit cell given as 3x3 array.
            used_indices(set): A set of indices for atoms that have already
                been used as seed atoms or as part of unit cells in the system.
            cell_index(tuple): Index of the given cell in the
                LinkedUnitCollection. Given relative to the initial seed atom.
            searched_cell_indices(set of tuples): A set of cell indices that have
                already been searched.

        Returns:
            np.ndarray: The new unit cell that should be used when expanding
                the search.
            np.ndarray: Indices of the atoms that should be used as new seed
                atoms.
            np.ndarray: Positions of the new seed atoms.
            np.ndarray: Indices of the cells corresponding to the new seed
                atoms.
        """
        new_seed_indices = []
        new_seed_pos = []
        new_cell_indices = []
        new_cell = np.array(old_cell)

        # Check that no seed index is handled twice
        if seed_index in used_points:
            return new_cell, new_seed_indices, new_seed_pos, new_cell_indices
        else:
            used_points.add(seed_index)

        orig_cell = system.get_cell()
        orig_pos = system.get_positions()

        # Filter out cells that have already been searched
        test_cell_indices = multipliers + cell_index
        valid_multipliers = []
        for i_cell_ind, cell_ind in enumerate(test_cell_indices):
            # If the cell in this index has already been handled, continue
            if tuple(cell_ind) in searched_cell_indices:
                continue
            valid_multipliers.append(i_cell_ind)
        multipliers = multipliers[valid_multipliers]
        dislocations = dislocations[valid_multipliers]
        test_cell_indices = test_cell_indices[valid_multipliers]

        if seed_index is not None:

            # Find out the atoms that match the seed_guesses in the original
            # system
            seed_guesses = seed_pos + dislocations
            pos_tolerances = self.get_scaled_position_tolerance(dislocations)
            matches, _, _, factors = matid.geometry.get_matches(
                system,
                seed_guesses,
                len(dislocations)*[seed_atomic_number],
                pos_tolerances,
                mic=True
            )
            for match, factor, seed_guess, multiplier, disloc, test_cell_index in zip(
                    matches,
                    factors,
                    seed_guesses,
                    multipliers,
                    dislocations,
                    test_cell_indices):
                multiplier_tuple = tuple(multiplier)

                # Save the position corresponding to a seed atom or a guess for
                # it. If a match was found that is not the original seed, use
                # it's position to update the cell. If the matched index is the
                # same as the original seed, check the factors array to decide
                # whether to use the guess or not.
                if match is not None:
                    if match != seed_index:
                        i_seed_pos = orig_pos[match]
                    else:
                        if (factor == 0).all():
                            i_seed_pos = seed_guess
                        else:
                            i_seed_pos = orig_pos[match]
                else:
                    i_seed_pos = seed_guess

                # Check if this index has already been used as a seed. The
                # used_seed_indices is needed so that the same atom cannot
                # become a seed point multiple times. This can otherwise
                # become a problem in e.g. random systems, or "looped"
                # structures.
                add = True

                if match is not None:

                    # Get the index of the matched cell. If the matched atom is
                    # already associated with a cell, use that. Otherwise
                    # create a new index for the cell according to the
                    # multiplier.
                    if match in index_cell_map:
                        target_cell = index_cell_map[match]
                    else:
                        target_cell = cell_index + multiplier
                        index_cell_map[match] = target_cell

                    # Add an edge to the search graph
                    search_graph.add_node(tuple(cell_index), index=seed_index)
                    search_graph.add_node(tuple(target_cell), index=match)
                    search_graph.add_edge(tuple(cell_index), tuple(target_cell), multiplier=multiplier)

                    if match in used_indices:
                        add = False

                if add:
                    new_seed_indices.append(match)
                    new_seed_pos.append(i_seed_pos)
                    new_cell_indices.append(test_cell_index)

                    if match is not None:
                        used_indices.add(match)

                # Store the cell basis vector
                for i in range(3):
                    basis_mult = [0, 0, 0]
                    basis_mult[i] = 1
                    basis_mult = tuple(basis_mult)
                    if multiplier_tuple == basis_mult:
                        if match is None:
                            i_basis = disloc
                        else:
                            temp = i_seed_pos + np.dot(factor, orig_cell)
                            i_basis = temp - seed_pos
                        new_cell[i, :] = i_basis

        #TODO: Calculate the average cell for this seed atom. The average cell
        # is then used in the next phase of the search for the neighbouring
        # cells.

        return new_cell, new_seed_indices, new_seed_pos, new_cell_indices



from __future__ import absolute_import, division, print_function, unicode_literals


class SystaxError(Exception):
    def __init__(self, message, value=None):
        self.value = value
        Exception.__init__(self, message)


class ClassificationError(SystaxError):
    """Indicates that there was an error in finding a surface.
    """
    pass


class CellNormalizationError(SystaxError):
    """For errors in finding the normalized cell.
    """
    pass



from __future__ import division, unicode_literals
import numpy as np


class Lattice(object):
    """
    A lattice object.  Essentially a matrix with conversion matrices. In
    general, it is assumed that length units are in Angstroms and angles are in
    degrees unless otherwise stated.
    """
    def __init__(self, matrix):
        """
        Create a lattice from any sequence of 9 numbers. Note that the sequence
        is assumed to be read one row at a time. Each row represents one
        lattice vector.

        Args:
            matrix: Sequence of numbers in any form. Examples of acceptable
                input.
                i) An actual numpy array.
                ii) [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
                iii) [1, 0, 0 , 0, 1, 0, 0, 0, 1]
                iv) (1, 0, 0, 0, 1, 0, 0, 0, 1)
                Each row should correspond to a lattice vector.
                E.g., [[10, 0, 0], [20, 10, 0], [0, 0, 30]] specifies a lattice
                with lattice vectors [10, 0, 0], [20, 10, 0] and [0, 0, 30].
        """
        m = np.array(matrix, dtype=np.float64).reshape((3, 3))
        lengths = np.sqrt(np.sum(m ** 2, axis=1))
        self._lengths = lengths
        self._matrix = m
        self._angles = None
        self._inv_matrix = None

    @property
    def matrix(self):
        """Copy of matrix representing the Lattice"""
        return np.copy(self._matrix)

    @property
    def inv_matrix(self):
        """
        Inverse of lattice matrix.
        """
        if self._inv_matrix is None:
            self._inv_matrix = np.linalg.inv(self._matrix)
        return self._inv_matrix

    def get_cartesian_coords(self, fractional_coords):
        """
        Returns the cartesian coordinates given fractional coordinates.

        Args:
            fractional_coords (3x1 array): Fractional coords.

        Returns:
            Cartesian coordinates
        """
        return np.dot(fractional_coords, self._matrix)

    def get_fractional_coords(self, cart_coords):
        """
        Returns the fractional coordinates given cartesian coordinates.

        Args:
            cart_coords (3x1 array): Cartesian coords.

        Returns:
            Fractional coordinates.
        """
        return np.dot(cart_coords, self.inv_matrix)

    @property
    def lengths(self):
        if self._lengths is None:
            lengths = np.linalg.norm(self._matrix, axis=1)
            self._lengths = lengths
        return self._lengths

    @property
    def angles(self):
        """
        Returns the angles (alpha, beta, gamma) of the lattice.
        """
        if self._angles is None:
            # Angles
            angles = np.zeros(3)
            for i in range(3):
                j = (i + 1) % 3
                k = (i + 2) % 3
                angles[i] = np.dot(
                    self._matrix[j],
                    self._matrix[k]) / (self.lengths[j] * self.lengths[k])
            angles = np.clip(angles, -1.0, 1.0)
            self._angles = np.arccos(angles) * 180. / np.pi
        return self._angles

    @property
    def abc(self):
        """
        Lengths of the lattice vectors, i.e. (a, b, c)
        """
        return tuple(self.lengths)

    @property
    def alpha(self):
        """
        Angle alpha of lattice in degrees.
        """
        return self._angles[0]

    @property
    def beta(self):
        """
        Angle beta of lattice in degrees.
        """
        return self._angles[1]

    @property
    def gamma(self):
        """
        Angle gamma of lattice in degrees.
        """
        return self._angles[2]

    @property
    def volume(self):
        """
        Volume of the unit cell.
        """
        m = self._matrix
        return abs(np.dot(np.cross(m[0], m[1]), m[2]))

    @property
    def lengths_and_angles(self):
        """
        Returns (lattice lengths, lattice angles).
        """
        return tuple(self.lengths), tuple(self.angles)

    def get_reciprocal_lattice(self):
        """
        Return the reciprocal lattice. Note that this is the standard
        reciprocal lattice used for solid state physics with a factor of 2 *
        pi. If you are looking for the crystallographic reciprocal lattice,
        use the reciprocal_lattice_crystallographic property.
        The property is lazily generated for efficiency.
        """
        try:
            return self._reciprocal_lattice
        except AttributeError:
            v = self.inv_matrix.T
            self._reciprocal_lattice = Lattice(v * 2 * np.pi)
            return self._reciprocal_lattice

    def get_reciprocal_lattice_crystallographic(self):
        """
        Returns the *crystallographic* reciprocal lattice, i.e., no factor of
        2 * pi.
        """
        return Lattice(self.get_reciprocal_lattice().matrix / (2 * np.pi))



import sys
from setuptools import setup, find_packages

# Check python version
if sys.version_info[:2] < (3, 0):
    raise RuntimeError("Python version >= 3.0 required.")

if __name__ == "__main__":
    setup(name="matid",
        version="0.5.7",
        url="https://singroup.github.io/matid/",
        description=(
            "MatID is a python package for identifying and analyzing atomistic "
            "systems based on their structure."
        ),
        long_description=(
            "MatID is a python package for identifying and analyzing atomistic "
            "systems based on their structure."
        ),
        author='Lauri Himanen',
        license="Apache License 2.0",
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Topic :: Scientific/Engineering :: Physics",
            "License :: OSI Approved :: Apache Software License",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3 :: Only",
        ],
        keywords='atoms structure materials science crystal symmetry',
        packages=find_packages(),
        install_requires=[
            "future",
            "numpy",
            "scipy",
            "ase",
            "spglib>=1.10.1",
            "sklearn",
            "networkx>=2.4",
            "chronic"
        ],
        python_requires=">=3.5",
    )



"""This module defines functions for deriving geometry related quantities from
a atomic system.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import math
import itertools
from collections import defaultdict

import numpy as np
from numpy.random import RandomState

from ase.data import covalent_radii
from ase import Atom, Atoms
import ase.geometry

from matid.data.element_data import get_covalent_radii
# from matid.exceptions import SystaxError
from matid.core.linkedunits import Substitution
import matid.geometry

from sklearn.cluster import DBSCAN

from scipy.spatial import Delaunay
import spglib


def get_nearest_atom(system, position, mic=True):
    """Finds the index of the atom nearest to the given position in the given
    system.

    Args:
        system(ase.Atoms): The system from which the atom is searched from.
        position(np.ndarray): The position to search.
        mic(boolean): Whether to use the minimum image convention on periodic
            systems.

    Returns:
        int: Index of the nearest atom in the given system.
    """
    positions = system.get_positions()
    pbc = system.get_pbc()
    cell = system.get_cell()
    distances = get_distance_matrix(position, positions, cell=cell, pbc=pbc, mic=mic)
    min_index = np.argmin(distances)

    return min_index


def get_dimensionality(
        system,
        cluster_threshold,
        dist_matrix_radii_mic_1x=None,
        return_clusters=False
    ):
    """Used to calculate the dimensionality of a system with a modified
    Topology Scaling Algorithm (TSA) (Michael Ashton, Joshua Paul, Susan B.
    Sinnott, and Richard G. Hennig Phys. Rev. Lett. 118, 106101).

    Args:
        system (ASE.Atoms): The system for which the dimensionality is
            evaluated.
        cluster_threshold(float): The epsilon value for the DBSCAN algorithm
            that is used to identify clusters within the unit cell.
        dist_matrix_radii_pbc (np.ndarray): A precalculated distance matrix
            that takes in to account the periodicity and has the covalent radii of
            the atoms already subtracted.
        return_clusters(boolean): Whether the clusters are returned

    Returns:
        int|none: The dimensionality of the system. If the dimensionality can't be
            evaluated because the system has multiple disconnected components in
            the original cell, None is returned.
        list: A list of clusters. Each entry in the list contains the indices
            of atoms in a cluster.
    """
    system_1x = system
    pbc = system_1x.get_pbc()
    num_1x = system_1x.get_atomic_numbers()
    cell_1x = system_1x.get_cell()

    # When calculating the displacement tensor we can ignore neighbouring
    # copies that are farther away than the clustering cutoff. The maximum mic
    # distance to allow is cluster_threshold + 2*max_radii
    radii = covalent_radii[num_1x]
    max_radii = radii.max()
    max_distance = cluster_threshold + 2*max_radii

    # 1x1x1 system
    if dist_matrix_radii_mic_1x is None:
        pos_1x = system.get_positions()
        displacements_mic_1x, dist_matrix_mic_1x = get_displacement_tensor(
            pos_1x,
            pos_1x,
            cell_1x,
            pbc,
            mic=True,
            max_distance=max_distance,
            return_distances=True
        )
        radii_1x = covalent_radii[num_1x]
        radii_matrix_1x = radii_1x[:, None] + radii_1x[None, :]
        dist_matrix_radii_mic_1x = dist_matrix_mic_1x - radii_matrix_1x

    # Check the number of clusters.
    clusters_1x = get_clusters(dist_matrix_radii_mic_1x, cluster_threshold, min_samples=1)
    n_clusters_1x = len(clusters_1x)

    # If the system consists of multiple components that are not connected
    # according to the clustering done here, then we cannot assess the
    # dimensionality.
    if n_clusters_1x > 1:
        dim = None
    else:
        # 2x2x2 system
        n_pbc = np.sum(pbc)
        if n_pbc > 0:

            repeats = np.array([1, 1, 1])
            repeats[pbc] = 2
            system_2x = system.repeat(repeats)
            pos_2x = system_2x.get_positions()
            cell_2x = system_2x.get_cell()
            num_2x = system_2x.get_atomic_numbers()
            displacements_mic_2x, dist_matrix_mic_2x = get_displacement_tensor(
                pos_2x,
                pos_2x,
                cell_2x,
                pbc,
                mic=True,
                max_distance=max_distance,
                return_distances=True
            )
            radii_2x = covalent_radii[num_2x]
            radii_matrix_2x = radii_2x[:, None] + radii_2x[None, :]
            dist_matrix_radii_mic_2x = dist_matrix_mic_2x - radii_matrix_2x

            # Check the number of clusters.
            clusters_2x = get_clusters(dist_matrix_radii_mic_2x, cluster_threshold, min_samples=1)
            n_clusters_2x = len(clusters_2x)

            # This is the analytic formula for the dimensionality based on cluster
            # size on a 2x supersystem.
            dim = int(n_pbc - math.log(n_clusters_2x, 2))
        else:
            dim = 0

    if return_clusters:
        return dim, clusters_1x
    else:
        return dim


def get_tetrahedra_decomposition(system, max_distance):
    """Used to decompose a series of 3D atomic coordinates into non-overlapping
    tetrahedron that together represent the atomic structure.

    """
    # TODO: Reuse the distance matrix from the original system, just filter out
    # the entries that do not belong to the valid basis. This way we can avoid
    # recalculating the distances.

    # TODO: Determine the distance matrix of the padded system during the
    # construction of the padded system. This should be doable by simply using
    # the internal displacements and the cell periodicity.

    class TetrahedraDecomposition():
        """A class that represents a collection of tetrahedron.
        """
        def __init__(self, delaunay, invalid_simplex_indices):
            self.delaunay = delaunay
            self.invalid_simplex_indices = invalid_simplex_indices

        def find_simplex(self, pos):
            """Used to find the index of the simplex in which the given
            position resides in.

            Args:
                pos(np.nadrray): Position for which to find the simplex

            Returns:
                int: Index of the simplex in which the point is in. Returns
                None if simplex was not found.
            """
            index = self.delaunay.find_simplex(pos).item()
            if index == -1:
                return None
            if index not in self.invalid_simplex_indices:
                return index
            return None

    cell = system.get_cell()
    pos = system.get_positions()
    num = system.get_atomic_numbers()

    # Calculate a matrix containing the combined radii for each pair of atoms
    # in an extended system
    radii = covalent_radii[num]
    radii_matrix = radii[:, None] + radii[None, :]

    displacements_finite = get_displacement_tensor(pos, pos)

    # In order for the decomposition to cover also the edges, we have to extend
    # the system to cover also into the adjacent periodic images. That is done
    # within this loop.
    tesselation_atoms = system.copy()
    multipliers = np.array(list(itertools.product((-1, 0, 1), repeat=3)))

    for mult in multipliers:
        if tuple(mult) != (0, 0, 0):
            disloc = np.dot(mult, cell)

            # If system is periodic in this direction, calculate the distance
            # between atoms in the periodically repeated images and choose
            # atoms from the copies that are within a certain range when tha
            # radii are taken into account
            disp = np.array(displacements_finite)
            disp += disloc
            dist = np.linalg.norm(disp, axis=2)
            dist_radii = dist - radii_matrix
            connectivity_mask = np.where(dist_radii < max_distance)

            pad_pos = pos[connectivity_mask[1]] + disp[connectivity_mask[0], connectivity_mask[1]]
            pad_num = num[connectivity_mask[1]]
            pad_atoms = Atoms(positions=pad_pos, symbols=pad_num)
            tesselation_atoms += pad_atoms

    # view(tesselation_atoms)
    tesselation_pos = tesselation_atoms.get_positions()

    # The QJ options makes sure that all positions are included in the
    # tesselation
    tri = Delaunay(tesselation_pos, qhull_options="QJ")

    # Remove invalid simplices from the surface until only valid simplices are
    # found on the surface
    distance_matrix = get_covalent_distances(tesselation_atoms, mic=False)
    simplices = np.array(tri.simplices)
    neighbors = np.array(tri.neighbors)
    end = False
    invalid_indices = set()
    surface_simplices = np.any(neighbors == -1, axis=1)
    original_indices = np.arange(len(simplices))
    surface_simplex_indices = set(original_indices[surface_simplices])

    while not end:

        # Remove the surface simplices that are too big. Also update the
        # neighbour list.
        too_big_simplex_indices = []
        for i_simplex in surface_simplex_indices:
            simplex = simplices[i_simplex]
            for i, j in itertools.combinations(simplex, 2):
                distance = distance_matrix[i, j]
                if distance >= max_distance:
                    too_big_simplex_indices.append(i_simplex)
                    break
        invalid_indices = invalid_indices | set(too_big_simplex_indices)
        if len(too_big_simplex_indices) == 0:
            end = True
        else:
            # Find the neighbours of the simplices that were removed
            mask = np.isin(neighbors, too_big_simplex_indices)
            affected = np.any(mask, axis=1)
            new_surface_simplex_indices = set(original_indices[affected])
            new_surface_simplex_indices -= invalid_indices
            surface_simplex_indices = new_surface_simplex_indices

    tetrahedras = TetrahedraDecomposition(tri, invalid_indices)
    return tetrahedras


def get_moments_of_inertia(system, weight=True):
    """Calculates geometric inertia tensor, i.e., inertia tensor but with
    all masses are set to 1.

    I_ij = sum_k m_k (delta_ij * r_k^2 - x_ki * x_kj)
    with r_k^2 = x_k1^2 + x_k2^2 x_k3^2

    Args:
        system(ASE Atoms): Atomic system.

    Returns:
        (np.ndarray, np.ndarray): The eigenvalues and eigenvectors of the
        geometric inertia tensor.
    """
    # Move the origin to the geometric center
    positions = system.get_positions()
    centroid = get_center_of_mass(system, weight)
    pos_shifted = positions - centroid

    # Calculate the geometric inertia tensor
    if weight:
        weights = system.get_masses()
    else:
        weights = np.ones((len(system)))
    x = pos_shifted[:, 0]
    y = pos_shifted[:, 1]
    z = pos_shifted[:, 2]
    I11 = np.sum(weights*(y**2 + z**2))
    I22 = np.sum(weights*(x**2 + z**2))
    I33 = np.sum(weights*(x**2 + y**2))
    I12 = np.sum(-weights * x * y)
    I13 = np.sum(-weights * x * z)
    I23 = np.sum(-weights * y * z)

    I = np.array([
        [I11, I12, I13],
        [I12, I22, I23],
        [I13, I23, I33]])

    evals, evecs = np.linalg.eigh(I)

    return evals, evecs


def get_center_of_mass(system):
    """Calculates the center of mass and also takes the periodicity of the
    system into account.

    The algorithm is replicated from:
    https://en.wikipedia.org/wiki/Center_of_mass#Systems_with_periodic_boundary_conditions

    Args:
        system(ase.Atoms): The system for which the center of mass is
            calculated.

    Returns:
        np.ndarray: The cartesian positions of the center of mass in the given
        system.
    """
    pbc = system.get_pbc()
    relative_positions = system.get_scaled_positions()
    masses = system.get_masses()
    total_mass = np.sum(masses)
    cell = system.get_cell()

    rel_com = np.zeros((3))
    for i_comp in range(3):
        i_pbc = pbc[i_comp]
        i_pos = relative_positions[:, i_comp]
        if i_pbc:
            theta = i_pos * 2*np.pi
            xi = np.cos(theta)*masses
            zeta = np.sin(theta)*masses

            xi_mean = np.mean(xi)
            zeta_mean = np.mean(zeta)

            mean_theta = np.arctan2(-zeta_mean, -xi_mean) + np.pi
            com_rel = mean_theta/(2*np.pi)
            rel_com[i_comp] = com_rel
        else:
            rel_com[i_comp] = np.sum(i_pos*masses)/total_mass

    com_cart = to_cartesian(rel_com, cell)

    return com_cart


def get_space_filling(system):
    """Calculates the ratio of vacuum to filled space by assuming covalent
    radii for the atoms.

    Args:
        system(ASE.Atoms): Atomic system.

    Returns:
        float: The ratio of occupied volume to the cell volume.
    """
    cell_volume = system.get_volume()
    atomic_numbers = system.get_atomic_numbers()
    occupied_volume = 0
    radii = get_covalent_radii(atomic_numbers)
    volumes = 4.0/3.0*np.pi*radii**3
    occupied_volume = np.sum(volumes)
    ratio = occupied_volume/cell_volume

    return ratio


def make_random_displacement(system, delta, rng=None):
    """Dislocate every atom in the given system in a random direction but by
    the same amount. Uses an internal random number generator to avoid touching
    the global numpy.random.seed()-function.

    Args:
        system(ASE.Atoms): The system for which the displacement are performed.
        delta(float): The magnitude of the displacements.
        rng(np.random.RandomState): Random number generator.
    """
    if rng is None:
        rng = RandomState()
    pos = system.get_positions()
    n_atoms = len(system)
    disloc = rng.rand(n_atoms, 3)
    disloc /= np.linalg.norm(disloc, axis=1)[:, None]
    disloc *= delta
    new_pos = pos + disloc
    system.set_positions(new_pos)


def get_extended_system(system, target_size):
    """Replicate the system in different directions to reach the given target
    size.

    Args:
        system (ase.Atoms): The original system.
        target_size (float): The target size for the extended system in
            angstroms.

    Returns:
        ase.Atoms: The extended system.
    """
    pbc = system.get_pbc()
    cell = system.get_cell()

    repetitions = np.array([1, 1, 1])
    for i, pbc in enumerate(pbc):
        # Only extend in the periodic dimensions
        basis = cell[i, :]
        if pbc:
            size = np.linalg.norm(basis)
            i_repetition = np.maximum(np.round(target_size/size), 1).astype(int)
            repetitions[i] = i_repetition

    extended_system = system.repeat(repetitions)

    return extended_system


def get_clusters(dist_matrix, threshold, min_samples=1):
    """Used to detect clusters with the DBSCAN algorithm.

    Args:
        dist_matrix(np.ndarray): The 2D distance matrix from which the clusters
            are calculated.
        threshold(float): The epsilon threshold value for the clustering
        min_samples(int): The minimum allowed cluster size.

    Returns:
        list: A list of clusters, where each cluster is a list of indices for
        the elements belonging to the cluster.
    """
    # As the distance have been normalized with respect to the covalent radiis,
    # the distance matrix may in some cases have negative values. This simply
    # means that the distance between two atoms is smaller than the sum of
    # their covalent radiis. The distance values are here clipped to
    # zero to avoid problems in the cluster detection. Another option would be
    # to normalize the distances so that unity would correspond to the sum of
    # the two radii.
    np.clip(dist_matrix, a_min=0, a_max=None, out=dist_matrix)

    # Detect clusters
    db = DBSCAN(eps=threshold, min_samples=min_samples, metric='precomputed', n_jobs=1)
    db.fit(dist_matrix)
    clusters = db.labels_

    # Make a list of the different clusters. All item with cluster = -1 will
    # get a separate group.
    cluster_groups = []
    group_map = defaultdict(list)
    for i_atom, i_clust in enumerate(clusters):
        if i_clust == -1:
            cluster_groups.append([i_atom])
        else:
            group_map[i_clust].append(i_atom)
    cluster_groups.extend(group_map.values())

    return cluster_groups


def get_covalent_distances(system, mic=True):
    """Returns a distance matrix where the covalent radii have been taken into
    account. Clips negative values to be zero.
    """
    if system is not None:
        if len(system) == 1:
            return np.array([[0]])

        # Calculate distance matrix with radii taken into account
        distance_matrix = system.get_all_distances(mic=mic)

    # Remove the radii from distances and clip out negative values
    numbers = system.get_atomic_numbers()
    radii = covalent_radii[numbers]
    radii_matrix = radii[None, :] + radii[:, None]
    distance_matrix -= radii_matrix
    np.clip(distance_matrix, 0, None, out=distance_matrix)

    return distance_matrix


def get_biggest_gap_indices(coordinates):
    """Given the list of coordinates for one axis, this function will find the
    maximum gap between them and return the index of the bottom and top
    coordinates. The bottom and top are defined as:

    ===       ===============    --->
        ^top    ^bot               ^axis direction
    """
    # Find the maximum vacuum gap for all basis vectors
    sorted_indices = np.argsort(coordinates)
    sorted_comp = coordinates[sorted_indices]
    rolled_comp = np.roll(sorted_comp, 1, axis=0)
    distances = sorted_comp - rolled_comp

    # The first distance is from first to last, so it needs to be
    # wrapped around
    distances[0] += 1

    # Find maximum gap
    bottom_index = sorted_indices[np.argmax(distances)]
    top_index = sorted_indices[np.argmax(distances)-1]

    return bottom_index, top_index


def get_dimensions(system, vacuum_gaps):
    """Given a system with vacuum gaps, calculate its dimensions in the
    directions with vacuum gaps by also taking into account the atomic radii.
    """
    orig_cell_lengths = np.linalg.norm(system.get_cell(), axis=1)

    # Create a repeated copy of the system. The repetition is needed in order
    # to get gaps to neighbouring cell copies in the periodic dimensions with a
    # vacuum gap
    sys = system.copy()
    sys = sys.repeat([2, 2, 2])

    dimensions = [None, None, None]
    numbers = sys.get_atomic_numbers()
    positions = sys.get_scaled_positions()
    radii = covalent_radii[numbers]
    cell_lengths = np.linalg.norm(sys.get_cell(), axis=1)
    radii_in_cell_basis = radii[:, None]/cell_lengths[None, :]

    for i_dim, vacuum_gap in enumerate(vacuum_gaps):
        if vacuum_gap:
            # Make a data structure containing the atom location information as
            # intervals from one side of the atom to the other in each
            # dimension.
            intervals = Intervals()
            for i_pos, pos in enumerate(positions[:, i_dim]):
                i_radii = radii_in_cell_basis[i_pos, i_dim]
                i_axis_start = pos - i_radii
                i_axis_end = pos + i_radii
                intervals.add_interval(i_axis_start, i_axis_end)

            # Calculate the maximum distance between atoms, when taking radius
            # into account
            gap = intervals.get_max_distance_between_intervals()
            gap = gap*cell_lengths[i_dim]
            dimensions[i_dim] = orig_cell_lengths[i_dim] - gap

    return dimensions


def get_wrapped_positions(scaled_pos, precision=1E-5):
    """Wrap the given relative positions so that each element in the array
    is within the half-closed interval [0, 1)

    By wrapping values near 1 to 0 we will have a consistent way of
    presenting systems.
    """
    scaled_pos %= 1

    abs_zero = np.absolute(scaled_pos)
    abs_unity = np.absolute(abs_zero-1)

    near_zero = np.where(abs_zero < precision)
    near_unity = np.where(abs_unity < precision)

    scaled_pos[near_unity] = 0
    scaled_pos[near_zero] = 0

    return scaled_pos


def get_distance_matrix(pos1, pos2, cell=None, pbc=None, mic=False, use_self_distance=True):
    """Calculates the distance matrix. If wrap_distances=True, calculates
    the matrix using periodic distances

    Args:
        pos1(np.ndarray): Array of cartesian positions
        pos2(np.ndarray): Array of cartesian positions
        cell(): The 3x3 cell of the system. Needed if minimum image convention
            enabled.
        pbc(): The periodic boundary conditions of the system. Needed if
            minimum image convention enabled.
        mic (bool): Whether to apply minimum image convention fo the distances,
            i.e. the distance to the closest periodic image is returned.
        use_self_distance(boolean): Determines how the distance from atom to
            itself is determined. If True, zero is returned, if false, the
            distance to the closest copy is returned.

    Returns:
        np.ndarray: A :math:`N_{atoms} \times N_{atoms}` matrix of distances.
    """
    disp_tensor = get_displacement_tensor(pos1, pos2, cell, pbc, mic, use_self_distance=use_self_distance)
    distance_matrix = np.linalg.norm(disp_tensor, axis=2)

    return distance_matrix


def get_displacement_tensor(
        pos1,
        pos2,
        cell=None,
        pbc=None,
        mic=False,
        max_distance=None,
        return_factors=False,
        return_distances=False,
        use_self_distance=True
    ):
    """Given an array of positions, calculates the 3D displacement tensor
    between the positions.

    The displacement tensor is a matrix where the entry A[i, j, :] is the
    vector pos1[i] - pos2[j], i.e. the vector from pos2 to pos1

    Args:
        pos1(np.ndarray): 2D array of positions
        pos2(np.ndarray): 2D array of positions
        cell(np.ndarray): Cell for taking into account the periodicity
        pbc(boolean or a list of booleans): Periodicity of the axes
        mic(boolean): Whether to return the displacement to the nearest
            periodic copy
        mic_copies(np.ndarray): The maximum number of periodic copies to
            consider in each direction. If not specified, the maximum possible
            number of copies is determined and used.
        use_self_distance(boolean): Determines how the distance from atom to
            itself is determined. If True, zero is returned, if false, the
            distance to the closest copy is returned.

    Returns:
        np.ndarray: 3D displacement tensor
        (optional) np.ndarray: The indices of the periodic copies in which the
            minimum image was found
        (optional) np.ndarray: The lengths of the displacement vectors
    """
    # Make 1D into 2D
    shape1 = pos1.shape
    shape2 = pos2.shape
    if len(shape1) == 1:
        n_cols1 = len(pos1)
        pos1 = np.reshape(pos1, (-1, n_cols1))
    if len(shape2) == 1:
        n_cols2 = len(pos2)
        pos2 = np.reshape(pos2, (-1, n_cols2))

    # Calculate the pairwise distance vectors
    disp_tensor = pos1[:, None, :] - pos2[None, :, :]
    tensor_shape = np.array(disp_tensor.shape)

    # If minimum image convention is used, get the corrected vectors
    if mic:
        flattened_disp = np.reshape(disp_tensor, (-1, 3))
        D, D_len, factors = find_mic(flattened_disp, cell, pbc, max_distance)
        disp_tensor = np.reshape(D, tensor_shape)
        if return_factors:
            factors = np.reshape(factors, tensor_shape)
        if return_distances:
            lengths = np.reshape(D_len, (tensor_shape[0], tensor_shape[1]))
    else:
        if return_factors:
            factors = np.zeros((tensor_shape[0], tensor_shape[1], 3))
        if return_distances:
            lengths = np.linalg.norm(disp_tensor, axis=2)

    if return_factors and return_distances:
        return disp_tensor, factors, lengths
    elif return_factors:
        return disp_tensor, factors
    elif return_distances:
        return disp_tensor, lengths
    else:
        return disp_tensor


def find_mic(D, cell, pbc, max_distance=None):
    """Finds the minimum-image representation of vector(s) D.

    Args:
        D(np.ndarray): A nx3 array of vectors.
        cell(np.ndarray): A valid ase Atoms cell definition.
        pbc(np.ndarray): A 3x1 boolean array for the periodic boundary
            conditions.
        mic_copies(np.ndarray): The maximum number of periodic copies to
            consider in each direction. If not specified, the maximum possible
            number of copies is determined and used.

    Returns:
        np.ndarray: The minimum image versions of the given vectors.
        np.ndarray: The shifts corresponding to the indices of the neighbouring
            cells in which the vectors were found.
    """
    cell = ase.geometry.complete_cell(cell)
    n_vectors = len(D)

    # Calculate the 4 unique unit cell diagonal lengths
    diags = np.sqrt((np.dot([[1, 1, 1],
                             [-1, 1, 1],
                             [1, -1, 1],
                             [-1, -1, 1],
                             ], cell)**2).sum(1))

    Dr = np.dot(D, np.linalg.inv(cell))

    # calculate 'mic' vectors (D) and lengths (D_len) using simple method
    # return mic vectors and lengths for only orthorhombic cells,
    # as the results may be wrong for non-orthorhombic cells
    if (max(diags) - min(diags)) / max(diags) < 1e-9:

        factors = np.floor(Dr + 0.5)*pbc
        D = np.dot(Dr - factors, cell)
        D_len = np.linalg.norm(D, axis=1)
        return D, D_len, factors
    # Non-orthorhombic cell
    else:
        D_len = np.linalg.norm(D, axis=1)

    # The cutoff radius is the longest direct distance between atoms
    # or half the longest lattice diagonal, whichever is smaller
    if max_distance is None:
        max_distance = max(D_len)
    mic_cutoff = min(max_distance, max(diags) / 2.)

    # Construct a list of translation vectors. For example, if we are
    # searching only the nearest images (27 total), tvecs will be a
    # 27x3 array of translation vectors.
    tvec_factors = get_neighbour_cells(cell, mic_cutoff, pbc)
    tvecs = np.dot(tvec_factors, cell)

    # Translate the direct displacement vectors by each translation
    # vector, and calculate the corresponding lengths.
    D_trans = tvecs[np.newaxis] + D[:, np.newaxis]
    D_trans_len = np.linalg.norm(D_trans, axis=2)

    # Find mic distances and corresponding vector(s) for each given pair
    # of atoms. For symmetrical systems, there may be more than one
    # translation vector corresponding to the MIC distance; this finds the
    # first one in D_trans_len.
    D_min_ind = D_trans_len.argmin(axis=1)
    D_min_len = D_trans_len[list(range(n_vectors)), D_min_ind]
    D_min = D_trans[list(range(n_vectors)), D_min_ind]
    factors = tvec_factors[D_min_ind, :]

    # Negate the factors because of the order of the displacement
    factors *= -1

    return D_min, D_min_len, factors


def get_neighbour_cells(cell, cutoff, pbc):
    """Given a cell and a cutoff, returns the indices of the copies of the cell
    which have to be searched in order to reach atom within the cutoff
    distance.

    The number of neighboring images to search in each direction is equal to
    the ceiling of the cutoff distance (defined above) divided by the length of
    the projection of the lattice vector onto its corresponding surface normal.
    a's surface normal vector is e.g.  b x c / (|b| |c|), so this projection is
    (a . (b x c)) / (|b| |c|).  The numerator is just the lattice volume, so
    this can be simplified to V / (|b| |c|). This is rewritten as V |a| / (|a|
    |b| |c|) for vectorization purposes.

    Args:

    Returns:
    """
    pbc = expand_pbc(pbc)
    latt_len = np.linalg.norm(cell, axis=1)
    V = abs(np.linalg.det(cell))
    mic_copies = pbc * np.array(np.ceil(cutoff * np.prod(latt_len) /
                            (V * latt_len)), dtype=int)
    n0 = range(-mic_copies[0], mic_copies[0] + 1)
    n1 = range(-mic_copies[1], mic_copies[1] + 1)
    n2 = range(-mic_copies[2], mic_copies[2] + 1)

    factors = cartesian((n0, n1, n2))

    return factors


def get_mic_vector(w, v, cell):
    """Used to calculate the minimum image version of a vector in the given
    cell.

    Args:
        rel_vector(np.ndarray): Relative vector in the cell basis:

    Returns:
        np.ndarray: The MIC version of the given vector
        np.ndarray: The shift that corresponds to the neighbouring cell in
            which the copy was found.
    """
    # Get the original shift
    dvw = w-v
    cart_vec = to_cartesian(dvw, cell)

    orig_shift = np.array([0, 0, 0])
    for i in range(0, 3):
        if dvw[i] < -0.5 or dvw[i] > 0.5:
            k = np.floor(dvw[i] + 0.5)
            dvw[i] -= k
            orig_shift[i] = -k

    # cart_vec = to_cartesian(dvw, cell)

    # Figure out in which block of 2x2x2 cells we are testing
    init_val = np.array([0, 0, 0])
    init_val[dvw > 0] = -1

    # Figure out the minimum vector in this block of 8 cells
    a_range = range(init_val[0], init_val[0]+2)
    b_range = range(init_val[1], init_val[1]+2)
    c_range = range(init_val[2], init_val[2]+2)
    multipliers = cartesian((a_range, b_range, c_range))

    vec_copies = cart_vec + np.dot(multipliers, cell)
    squared_len = np.sum(vec_copies**2)
    min_index = np.argmin(squared_len)
    min_shift = multipliers[min_index]

    shift = orig_shift + min_shift

    addition = np.dot(shift, cell)
    result = cart_vec + addition

    return result, shift


def expand_pbc(pbc):
    """Used to expand a pbc definition into an array of three booleans.

    Args:
        pbc(boolean or a list of booleans): The periodicity of the cell. This
            can be any of the values that is also supprted by ASE, namely: a
            boolean or a list of three booleans.

    Returns:
        np.ndarray of booleans: The periodicity expanded as an explicit list of
        three boolean values.
    """
    if pbc is True:
        new_pbc = [True, True, True]
    elif pbc is False:
        new_pbc = [False, False, False]
    elif len(pbc) == 3:
        new_pbc = pbc
    else:
        raise ValueError(
            "Could not interpret the given periodic boundary conditions: '{}'"
            .format(pbc)
        )

    return np.array(new_pbc)


def change_basis(positions, basis, offset=None):
    """Transform the given cartesian coordinates to a basis that is defined by
    the given basis and origin offset.

    Args:
        positions(np.ndarray): Positions in cartesian coordinates.
        basis(np.ndarray): Basis to which to transform.
        offset(np.ndarray): Offset of the origins. A vector from the old basis
            origin to the new basis origin.
    Returns:
        np.ndarray: Relative positions in the new basis
    """

    if offset is not None:
        positions -= offset
    new_basis_inverse = np.linalg.inv(basis.T)
    pos_prime = np.dot(positions, new_basis_inverse.T)

    return pos_prime


def get_positions_within_basis(
        system,
        basis,
        origin,
        tolerance_low,
        tolerance_high,
        mask=[True, True, True],
        pbc=True
    ):
    """Used to return the indices of positions that are inside a certain basis.
    Also takes periodic boundaries into account.

    Args:
        system(ASE.Atoms): System from which the positions are searched.
        basis(np.ndarray): New basis vectors.
        origin(np.ndarray): New origin of the basis in cartesian coordinates.
        tolerance(float): The tolerance for the end points of the cell.
        mask(sequence of bool): Mask for selecting the basis's to consider.
        pbc(sequence of bool): The periodicity of the system.

    Returns:
        sequence of int: Indices of the atoms within this cell in the given
            system.
        np.ndarray: Relative positions of the found atoms.
        np.ndarray: The index of the periodic copy in which the position was
            found.
    """
    # If the search extend beyound the cell boundary and periodic boundaries
    # allow, we must divide the search area into multiple regions

    # Transform positions into the new basis
    cart_pos = system.get_positions()
    orig_cell = system.get_cell()
    pbc = expand_pbc(pbc)

    # We need to expand the system in different directions to get the

    # See if the new positions extend beyound the boundaries. The original
    # simulation cell is always convex, so we can just check the corners of
    # unit cell defined by the basis
    max_a = origin + basis[0, :]
    max_b = origin + basis[1, :]
    max_c = origin + basis[1, :]
    max_ab = origin + basis[0, :] + basis[1, :]
    max_ac = origin + basis[0, :] + basis[2, :]
    max_bc = origin + basis[1, :] + basis[2, :]
    max_abc = origin + basis[0, :] + basis[1, :] + basis[2, :]
    vectors = np.array((max_a, max_b, max_c, max_ab, max_ac, max_bc, max_abc))
    rel_vectors = to_scaled(orig_cell, vectors, wrap=False, pbc=system.get_pbc())
    factors = np.floor(rel_vectors).astype(int)
    min_factors = np.min(factors, axis=0)
    max_factors = np.max(factors, axis=0)
    a_range = range(min_factors[0], max_factors[0]+1)
    b_range = range(min_factors[1], max_factors[1]+1)
    c_range = range(min_factors[2], max_factors[2]+1)
    factors = matid.geometry.cartesian((a_range, b_range, c_range))

    directions = []
    for factor in factors:
        a_per = factor[0]
        b_per = factor[1]
        c_per = factor[2]
        allow = True
        if a_per != 0 and not pbc[0]:
            allow = False
        if b_per != 0 and not pbc[1]:
            allow = False
        if c_per != 0 and not pbc[2]:
            allow = False
        if allow:
            directions.append(factor)

    # If the new cell is overflowing beyound the boundaries of the original
    # system, we have to also check the periodic copies.
    indices = []
    a_prec_low, b_prec_low, c_prec_low = tolerance_low/np.linalg.norm(basis, axis=1)
    a_prec_high, b_prec_high, c_prec_high = tolerance_high/np.linalg.norm(basis, axis=1)
    orig_basis = system.get_cell()
    cell_pos = []
    factors = []
    for i_dir in directions:

        vec_new_cart = cart_pos + np.dot(i_dir, orig_basis)
        vec_new_rel = change_basis(vec_new_cart - origin, basis)

        # If no positions are defined, find the atoms within the cell
        for i_pos, pos in enumerate(vec_new_rel):
            if mask[0]:
                x = 0 - a_prec_low <= pos[0] <= 1 - a_prec_high
            else:
                x = True
            if mask[1]:
                y = 0 - b_prec_low <= pos[1] <= 1 - b_prec_high
            else:
                y = True
            if mask[2]:
                z = 0 - c_prec_low <= pos[2] <= 1 - c_prec_high
            else:
                z = True

            if x and y and z:
                indices.append(i_pos)
                cell_pos.append(pos)
                factors.append(i_dir)

    cell_pos = np.array(cell_pos)
    indices = np.array(indices)
    factors = np.array(factors)

    return indices, cell_pos, factors


def get_matches(
        system,
        positions,
        numbers,
        tolerances,
        mic=True
    ):
    """Given a system and a list of cartesian positions and atomic numbers,
    returns a list of indices for the atoms corresponding to the given
    positions with some tolerance.

    Args:
        system(ASE.Atoms): System where to search the positions
        positions(np.ndarray): Positions to match in the system.
        tolerances(np.ndarray): Maximum allowed distance for each vector that
            is allowed for a match in position.
        mic(boolean): Whether to find the minimum image copy.

    Returns:
        np.ndarray: indices of matched atoms
        list: list of substitutions
        list: list of vacancies
        np.ndarray: for each searched position, an integer array representing
            the number of the periodic copy where the match was found.
    """
    orig_num = system.get_atomic_numbers()
    orig_pos = system.get_positions()
    cell = system.get_cell()
    pbc = system.get_pbc()
    pbc = expand_pbc(pbc)
    scaled_pos2 = to_scaled(cell, positions, wrap=False)

    disp_tensor, factors, dist_matrix = get_displacement_tensor(
        positions,
        orig_pos,
        cell,
        pbc,
        mic=mic,
        max_distance=tolerances.max(),
        return_factors=True,
        return_distances=True
    )

    # Find the closest atom within the tolerance and with the required atomic
    # number, or if not found, get the closest atom that is within the
    # tolerance
    best_matches = []
    best_substitutions = []
    for i_atom, i in enumerate(dist_matrix):
        near_mask = i <= tolerances[i_atom]
        element_mask = orig_num == numbers[i_atom]
        combined_mask = near_mask & element_mask
        possible_indices = np.where(combined_mask)[0]
        if len(possible_indices) != 0:
            min_dist_index = np.argmin(i[combined_mask])
            best_index = possible_indices[min_dist_index]
            best_matches.append(best_index)
            best_substitutions.append(None)
        elif near_mask.any():
            best_matches.append(None)
            near_indices = np.where(near_mask)[0]
            nearest_index = np.argmin(i[near_mask])
            best_substitutions.append(near_indices[nearest_index])
        else:
            best_matches.append(None)
            best_substitutions.append(None)

    # min_ind = np.argmin(dist_matrix, axis=1)
    matches = []
    substitutions = []
    vacancies = []
    copy_indices = []

    for i, (i_match, i_subst) in enumerate(zip(best_matches, best_substitutions)):

        match = None
        copy = None
        subst = None
        b_num = numbers[i]

        if i_match is not None:
            ind = i_match
            match = ind

            # If a match was found the factor is reported based on the
            # displacement tensor
            i_move = factors[i][ind]
            copy = i_move
        elif i_subst is not None:
            ind = i_subst
            a_num = orig_num[ind]

            # Wrap the substitute position
            subst_pos_cart = orig_pos[ind]
            subst = Substitution(ind, subst_pos_cart, b_num, a_num)

            # If a match was found the factor is reported based on the
            # displacement tensor
            i_move = factors[i][ind]
            copy = i_move
        else:
            vacancies.append(Atom(b_num, position=positions[i]))

            # If no match was found, the factor is reported from the scaled
            # positions
            copy = np.floor(scaled_pos2[i])

        substitutions.append(subst)
        matches.append(match)
        copy_indices.append(copy)

    return matches, substitutions, vacancies, copy_indices


def to_scaled(cell, positions, wrap=False, pbc=False):
    """Used to transform a set of positions to the basis defined by the
    cell of this system.

    Args:
        system(ASE.Atoms): Reference system.
        positions (numpy.ndarray): The positions to scale
        wrap (numpy.ndarray): Whether the positions should be wrapped
            inside the cell.

    Returns:
        numpy.ndarray: The scaled positions
    """
    # Force 1D to 2D
    if len(positions.shape) == 1:
        positions = positions[None, :]
    pbc = expand_pbc(pbc)
    fractional = np.linalg.solve(
        cell.T,
        positions.T).T

    if wrap:
        for i, periodic in enumerate(pbc):
            if periodic:
                fractional[:, i] %= 1.0

    return fractional


def to_cartesian(cell, scaled_positions, wrap=False, pbc=False):
    """Used to transofrm a set of relative positions to the cartesian basis
    defined by the cell of this system.

    Args:
        system (ASE.Atoms): Reference system.
        positions (numpy.ndarray): The positions to scale
        wrap (numpy.ndarray): Whether the positions should be wrapped
            inside the cell.

    Returns:
        numpy.ndarray: The cartesian positions
    """
    pbc = expand_pbc(pbc)
    if wrap:
        for i, periodic in enumerate(pbc):
            if periodic:
                scaled_positions[:, i] %= 1.0

    cartesian_positions = np.dot(scaled_positions, cell)
    return cartesian_positions


def translate(system, translation, relative=False):
    """Translates the positions by the given translation.

    Args:
        translation (1x3 numpy.array): The translation to apply.
        relative (bool): True if given translation is relative to cell
            vectors.
    """
    if relative:
        rel_pos = system.get_scaled_positions()
        rel_pos += translation
        system.set_scaled_positions(rel_pos)
    else:
        cart_pos = system.get_positions()
        cart_pos += translation
        system.set_positions(cart_pos)


def get_closest_direction(vec, directions, normalized=False):
    """Used to return the direction that is most parallel to a given one.

    Args:

    Returns:
    """
    if not normalized:
        directions = directions/np.linalg.norm(directions, axis=1)
    dots = np.abs(np.dot(vec, directions.T))
    index = np.argmax(dots)

    return index


class Intervals(object):
    """Handles list of intervals.

    This class allows sorting and adding up of intervals and taking into
    account if they overlap.
    """
    def __init__(self, intervals=None):
        """Args:
            intervals: List of intervals that are added.
        """
        self._intervals = []
        self._merged_intervals = []
        self._merged_intervals_need_update = True
        if intervals is not None:
            self.add_intervals(intervals)

    def _add_up(self, intervals):
        """Add up the length of intervals.

        Argument:
            intervals: List of intervals that are added up.

        Returns:
            Result of addition.
        """
        if len(intervals) < 1:
            return None
        result = 0.
        for interval in intervals:
            result += abs(interval[1] - interval[0])
        return result

    def add_interval(self, a, b):
        """Add one interval.

        Args:
            a, b: Start and end of interval. The order does not matter.
        """
        self._intervals.append((min(a, b), max(a, b)))
        self._merged_intervals_need_update = True

    def add_intervals(self, intervals):
        """Add list of intervals.

        Args:
            intervals: List of intervals that are added.
        """
        for interval in intervals:
            if len(interval) == 2:
                self.add_interval(interval[0], interval[1])
            else:
                raise ValueError("Intervals must be tuples of length 2!")

    def set_intervals(self, intervals):
        """Set list of intervals.

        Args:
            intervals: List of intervals that are set.
        """
        self._intervals = []
        self.add_intervals(intervals)

    def remove_interval(self, i):
        """Remove one interval.

        Args:
            i: Index of interval that is removed.
        """
        try:
            del self._intervals[i]
            self._merged_intervals_need_update = True
        except IndexError:
            pass

    def get_intervals(self):
        """Returns the intervals.
        """
        return self._intervals

    def get_intervals_sorted_by_start(self):
        """Returns list with intervals ordered by their start.
        """
        return sorted(self._intervals, key=lambda x: x[0])

    def get_intervals_sorted_by_end(self):
        """Returns list with intervals ordered by their end.
        """
        return sorted(self._intervals, key=lambda x: x[1])

    def get_merged_intervals(self):
        """Returns list of merged intervals so that they do not overlap anymore.
        """
        if self._merged_intervals_need_update:
            if len(self._intervals) < 1:
                return self._intervals
            # sort intervals in list by their start
            sorted_by_start = self.get_intervals_sorted_by_start()
            # add first interval
            merged = [sorted_by_start[0]]
            # start from second interval
            for current in sorted_by_start[1:]:
                previous = merged[-1]
                # new interval if not current and previous are not overlapping
                if previous[1] < current[0]:
                    merged.append(current)
                # merge if current and previous are overlapping and if end of previous is expanded by end of current
                elif previous[1] < current[1]:
                    merged[-1] = (previous[0], current[1])
            self._merged_intervals = merged
            self._merged_intervals_need_update = False
        return self._merged_intervals

    def get_max_distance_between_intervals(self):
        """Returns the maximum distance between the intervals while accounting for overlap.
        """
        if len(self._intervals) < 2:
            return None
        merged_intervals = self.get_merged_intervals()
        distances = []
        if len(merged_intervals) == 1:
            return 0.0
        for i in range(len(merged_intervals) - 1):
            distances.append(abs(merged_intervals[i + 1][0] - merged_intervals[i][1]))
        return max(distances)

    def add_up_intervals(self):
        """Returns the added up lengths of intervals without accounting for overlap.
        """
        return self._add_up(self._intervals)

    def add_up_merged_intervals(self):
        """Returns the added up lengths of merged intervals in order to account for overlap.
        """
        return self._add_up(self.get_merged_intervals())


def get_thickness(system, axis):
    """Get the thickness of an atomic system along a basis vector direction.

    Args:
        system(ase.Atoms): The system from which the thickness is evaluated.
        axis(int): The index of the unit cell basis in which direction the
            thickness is evaluated.

    Returns:
        float: The thickness of the system as measured from the center of the
        topmost atom to the center of lowermost atom.
    """
    pos = system.get_scaled_positions()[:, axis]
    pos_min = pos.min()
    pos_max = pos.max()
    basis = system.get_cell()[axis, :]
    thickness = (pos_max - pos_min)*np.linalg.norm(basis)

    return thickness


def get_minimized_cell(system, axis, min_size):
    """Used to resize the cell in the given cell direction so that all atoms
    are within the cell.

    Args:
        system(ase.Atoms): The system to minimize. Must have a cell with
            non-zero volume.
        axis(int): Index of the axis to minimize.
        axis(min_size): The minimum size for the cell in the given direction.

    Returns:
        ase.Atoms: The new minimized system.
    """
    # Grow the cell to fit all atoms
    rel_pos = system.get_scaled_positions(wrap=False)
    num = system.get_atomic_numbers()
    pbc = system.get_pbc()
    basis = system.get_cell()
    c = basis[axis, :]
    c_length = np.linalg.norm(c)
    c_norm = c/c_length
    c_comp = rel_pos[:, axis]
    # print(c_comp)

    min_index = np.argmin(c_comp, axis=0)
    max_index = np.argmax(c_comp, axis=0)

    pos_min_rel = np.zeros((3))
    pos_min_rel[axis] = c_comp[min_index]
    pos_max_rel = np.zeros((3))
    pos_max_rel[axis] = c_comp[max_index]

    pos_min_cart = matid.geometry.to_cartesian(basis, pos_min_rel)
    pos_max_cart = matid.geometry.to_cartesian(basis, pos_max_rel)
    c_real_cart = pos_max_cart-pos_min_cart
    c_size = np.linalg.norm(c_real_cart)

    # We demand a minimum size for the c-vector even if the system seems to
    # be purely 2-dimensional. This is done because the 3D-space cannot be
    # searched properly if one dimension is flat.
    if c_size < min_size:
        c_inflated_cart = min_size*c_norm
        c_new_cart = c_inflated_cart
    else:
        c_new_cart = c_real_cart
    new_basis = np.array(basis)
    new_basis[axis, :] = c_new_cart

    new_scaled_pos = rel_pos - pos_min_rel
    new_cart_test = matid.geometry.to_cartesian(basis, new_scaled_pos)
    new_scaled_pos = matid.geometry.to_scaled(new_basis, new_cart_test)

    # Translate the system to be in the middle if size is smaller than the
    # given minimum size
    if c_size < min_size:
        offset_cart = (c_real_cart-c_inflated_cart)/2
        offset_rel = matid.geometry.to_scaled(new_basis, offset_cart)
        new_scaled_pos -= offset_rel

    # Create translated system
    minimized_system = Atoms(
        cell=new_basis,
        scaled_positions=new_scaled_pos,
        symbols=num,
        pbc=pbc
    )

    return minimized_system


def cartesian(arrays, out=None):
    """
    Generate a cartesian product of input arrays.

    Args:
        arrays(sequence of arrays): The arrays from which the product is
            created.
        out(ndarray): Array to place the cartesian product in.

    Returns:
        ndarray: 2-D array of shape (M, len(arrays)) containing cartesian
        products formed of input arrays.

    Example:
    --------
    >>> cartesian(([1, 2, 3], [4, 5], [6, 7]))
    array([[1, 4, 6],
           [1, 4, 7],
           [1, 5, 6],
           [1, 5, 7],
           [2, 4, 6],
           [2, 4, 7],
           [2, 5, 6],
           [2, 5, 7],
           [3, 4, 6],
           [3, 4, 7],
           [3, 5, 6],
           [3, 5, 7]])
    """

    arrays = [np.asarray(x) for x in arrays]
    dtype = arrays[0].dtype

    n = np.prod([x.size for x in arrays])
    if out is None:
        out = np.zeros([n, len(arrays)], dtype=dtype)

    m = int(n / arrays[0].size)
    out[:, 0] = np.repeat(arrays[0], m)
    if arrays[1:]:
        cartesian(arrays[1:], out=out[0:m, 1:])
        for j in range(1, arrays[0].size):
            out[j*m:(j+1)*m, 1:] = out[0:m, 1:]
    return out


def get_crystallinity(symmetry_analyser):
    """Quantifies the crystallinity of the structure as a ratio of symmetries
    per number of unique atoms in primitive cell. This metric can be used to
    distinguish between amorphous and 'regular' crystals.

    The number of symmetry operations corresponds to the symmetry operations
    corresponding to the hall number of the structure. The symmetry operations
    as given by spglib.get_symmetry() are specific to the original structure,
    and they have not been reduced to the symmetries of the space group.

    Args:
        symmetry_analyser (): A SymmetryAnalyzer object that has been
            initialized with an atomic structure.

    Returns:
        (float) A ratio of symmetries per unique atoms in the primitive cell.
    """
    # Get the number of equivalent atoms in the primitive cell.
    n_unique_atoms_prim = len(symmetry_analyser.get_equivalent_atoms_primitive())

    hall_number = symmetry_analyser.get_hall_number()
    sym_ops = spglib.get_symmetry_from_database(hall_number)
    n_symmetries = len(sym_ops["rotations"])

    ratio = n_symmetries/float(n_unique_atoms_prim)

    return ratio


# def get_surface_normal_direction(system):
    # """Used to estimate a normal vector for a 2D like structure.

    # Args:
        # system (ase.Atoms): The system to examine.

    # Returns:
        # np.ndarray: The estimated surface normal vector
    # """
    # repeated = get_extended_system(system, 15)
    # # vectors = system.get_cell()

    # # Get the eigenvalues and eigenvectors of the moment of inertia tensor
    # val, vec = get_moments_of_inertia(repeated)
    # sorted_indices = np.argsort(val)
    # val = val[sorted_indices]
    # vec = vec[sorted_indices]

    # # If the moment of inertia is not significantly bigger in one
    # # direction, then the system cannot be described as a surface.
    # moment_limit = 1.5
    # if val[-1] < moment_limit*val[0] and val[-1] < moment_limit*val[1]:
        # raise ValueError(
            # "The given system could not be identified as a surface. Make"
            # " sure that you provide a surface system with a sufficient"
            # " vacuum gap between the layers (at least ~8 angstroms of vacuum"
            # " between layers.)"
        # )

    # # The biggest component is the orhogonal one
    # orthogonal_dir = vec[-1]

    # return orthogonal_dir

    # # Find out the cell direction that corresponds to the orthogonal one
    # # cell = repeated.get_cell()
    # # dots = np.abs(np.dot(orthogonal_dir, vectors.T))
    # # orthogonal_vector_index = np.argmax(dots)
    # # orthogonal_vector = vectors[orthogonal_vector_index]
    # # orthogonal_dir = orthogonal_vector/np.linalg.norm(orthogonal_vector)

    # return orthogonal_dir
